import sys
import os
import signal
import importlib
import psutil

from threading import Thread
from pkg_resources import resource_string, resource_filename
from functools import cmp_to_key
import locale
from PyQt5.QtCore import QThread

from cardholder.cardholder import CardHolder
from cardholder.cardholder import Card
from cardholder.cardholder import CollectCardsThread

from akoteka.gui.pyqt_import import *
from akoteka.gui.card_panel import CardPanel
from akoteka.gui.configuration_dialog import ConfigurationDialog

from akoteka.accessories import collect_cards
from akoteka.accessories import filter_key
from akoteka.accessories import clearLayout
from akoteka.accessories import FlowLayout
from akoteka.accessories import play_media



from akoteka.constants import *
from akoteka.setup.setup import getSetupIni

from akoteka.handle_property import _
from akoteka.handle_property import re_read_config_ini
from akoteka.handle_property import config_ini
from akoteka.handle_property import get_config_ini

# ================================
#
# Control Buttons Holder
#
# Contains:
#           Back Button
#           Fast Search Button
#           Advanced Search Button
# ================================
class ControlButtonsHolder(QWidget):
    def __init__(self, control_panel):
        super().__init__()
       
        self.control_panel = control_panel
        
        self_layout = QHBoxLayout(self)
        self.setLayout(self_layout)
        
        self_layout.setContentsMargins(0, 0, 0, 0)
        self_layout.setSpacing(5)
    
        # -------------
        #
        # Back Button
        #
        # -------------     
        self.back_button_method = None
        back_button = QPushButton()
        back_button.setFocusPolicy(Qt.NoFocus)
        back_button.clicked.connect(self.back_button_on_click)
        
        back_button.setIcon( QIcon( resource_filename(__name__,os.path.join("img", IMG_BACK_BUTTON)) ))
        back_button.setIconSize(QSize(32,32))
        back_button.setCursor(QCursor(Qt.PointingHandCursor))
        back_button.setStyleSheet("background:transparent; border:none") 

        # Back button on the left
        self_layout.addWidget( back_button )

        self_layout.addStretch(1) 
        
        
        # ================================================
        # ================================================        
        
        # -------------------
        #
        # Play Sequentially Button
        #
        # -------------------

        self.play_continously_button = QPushButton()
        self.play_continously_button.setFocusPolicy(Qt.NoFocus)
        self.play_continously_button.clicked.connect(self.play_continously_button_on_click)
        
        play_continously_icon = QIcon()
        play_continously_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_PLAY_CONTINOUSLY_BUTTON_ON)) ), QIcon.Normal, QIcon.On)
#        play_continously_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_PLAY_CONTINOUSLY_BUTTON_OFF)) ), QIcon.Normal, QIcon.Off)
        self.play_continously_button.setIcon( play_continously_icon )
        self.play_continously_button.setIconSize(QSize(25,25))
        self.play_continously_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.play_continously_button.setStyleSheet("background:transparent; border:none") 
        self_layout.addWidget( self.play_continously_button )

        # -------------------
        #
        # Stop Button
        #
        # -------------------
        self.stop_continously_button = QPushButton()
        self.stop_continously_button.setFocusPolicy(Qt.NoFocus)
        self.stop_continously_button.clicked.connect(self.stop_continously_button_on_click)
        
        stop_continously_icon = QIcon()
        stop_continously_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_STOP_CONTINOUSLY_BUTTON_ON)) ), QIcon.Normal, QIcon.On)
#        stop_continously_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_STOP_CONTINOUSLY_BUTTON_OFF)) ), QIcon.Normal, QIcon.Off)
        self.stop_continously_button.setIcon( stop_continously_icon )
        self.stop_continously_button.setIconSize(QSize(25,25))
        self.stop_continously_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.stop_continously_button.setStyleSheet("background:transparent; border:none") 
        self_layout.addWidget( self.stop_continously_button )
        
        # ----------------------------------
        #
        # Playing Continously list drop-down
        #
        # ----------------------------------

        self.dropdown_play_continously = QComboBox(self)
        self.dropdown_play_continously.setFocusPolicy(Qt.NoFocus)
        self.dropdown_play_continously.setEditable(True)
        
#        self.dropdown.currentIndexChanged.connect(self.current_index_changed)

        style_box = '''
            QComboBox { 
                max-width: 200px; min-width: 300px; border: 1px solid gray; border-radius: 5px;
            }
        '''
        style_drop_down ='''
            QComboBox QAbstractItemView::item { 
                color: red;
                max-height: 15px;
            }
        '''            
      
        self.dropdown_play_continously.setStyleSheet(style_box + style_drop_down)
        #self.dropdown_play_continously.addItem("")
        self_layout.addWidget( self.dropdown_play_continously )

               
        self_layout.addStretch(1) 
        
        # ================================================
        # ================================================        
                
        # -------------------
        #
        # Fast Search Button
        #
        # -------------------
        self.fast_search_button = QPushButton()
        self.fast_search_button.setFocusPolicy(Qt.NoFocus)
        self.fast_search_button.setCheckable(True)
        self.fast_search_button.setAutoExclusive(False)
        self.fast_search_button.toggled.connect(self.fast_search_button_on_click)
        
        fast_search_icon = QIcon()
        fast_search_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_FAST_SEARCH_BUTTON_ON)) ), QIcon.Normal, QIcon.On)
        fast_search_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_FAST_SEARCH_BUTTON_OFF)) ), QIcon.Normal, QIcon.Off)
        self.fast_search_button.setIcon( fast_search_icon )
        self.fast_search_button.setIconSize(QSize(25,25))
        self.fast_search_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.fast_search_button.setStyleSheet("background:transparent; border:none") 
        self_layout.addWidget( self.fast_search_button )
        
        # -------------------
        #
        # Advanced Search Button
        #
        # -------------------
        self.advanced_search_button = QPushButton()
        self.advanced_search_button.setFocusPolicy(Qt.NoFocus)
        self.advanced_search_button.setCheckable(True)
        self.advanced_search_button.setAutoExclusive(False)
        self.advanced_search_button.toggled.connect(self.advanced_search_button_on_click)
        
        advanced_search_icon = QIcon()
        advanced_search_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_ADVANCED_SEARCH_BUTTON_ON)) ), QIcon.Normal, QIcon.On)
        advanced_search_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_ADVANCED_SEARCH_BUTTON_OFF)) ), QIcon.Normal, QIcon.Off)
        self.advanced_search_button.setIcon( advanced_search_icon )
        self.advanced_search_button.setIconSize(QSize(25,25))
        self.advanced_search_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.advanced_search_button.setStyleSheet("background:transparent; border:none") 
        self_layout.addWidget( self.advanced_search_button )
        
        # ================================================
        # ================================================
                        
        # -------------------
        #
        # Config Button
        #
        # -------------------
        self.config_button = QPushButton()
        self.config_button.setFocusPolicy(Qt.NoFocus)
        self.config_button.setCheckable(False)
        self.config_button.clicked.connect(self.config_button_on_click)
        
        config_icon = QIcon()
        config_icon.addPixmap(QPixmap( resource_filename(__name__,os.path.join("img", IMG_CONFIG_BUTTON)) ), QIcon.Normal, QIcon.On)
        self.config_button.setIcon( config_icon )
        self.config_button.setIconSize(QSize(25,25))
        self.config_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.config_button.setStyleSheet("background:transparent; border:none") 
        self_layout.addWidget( self.config_button )   
        
        self.enableSearchIcons(False)
        self.disablePlayStopContinously()
        
    # ======================================================
    
    def clear_play_continously_elements(self):
        self.dropdown_play_continously.clear()
        
    def add_play_continously_separator(self):
        self.dropdown_play_continously.insertSeparator(self.dropdown_play_continously.__len__())        
        
    def add_play_continously_element(self, title, path):
        self.dropdown_play_continously.addItem(title, path)        
        
    def get_play_continously_selected_path(self):
        return self.dropdown_play_continously.itemData( self.dropdown_play_continously.currentIndex() )

    def get_play_continously_path_by_index(self, index):
        return self.dropdown_play_continously.itemData( index )

    def get_play_continously_selected_title(self):
        return self.dropdown_play_continously.itemText( self.dropdown_play_continously.currentIndex() )    
    
    def get_play_continously_title_by_index(self, index):
        return self.dropdown_play_continously.itemText( index )
    
    def get_play_continously_selected_index(self):
        return self.dropdown_play_continously.currentIndex()

    def get_play_continously_last_index(self):
        return self.dropdown_play_continously.count() - 1
    
    def select_play_continously_element_by_index(self, index):
        self.dropdown_play_continously.setCurrentIndex(index)
        
    # ======================================================
            
    def disablePlayStopContinously(self):
        self.play_continously_button.setEnabled(False)
        self.stop_continously_button.setEnabled(False)
        self.dropdown_play_continously.setEnabled(False)
        
    def enablePlayContinously(self, enabled):
        self.play_continously_button.setEnabled(enabled)
        self.stop_continously_button.setEnabled(not enabled)
        self.dropdown_play_continously.setEnabled(enabled)
    
    # =====================================================
         
    def enableSearchIcons(self, enabled):
        if self.advanced_search_button.isChecked():
            self.advanced_search_button.setChecked(False)

        if self.fast_search_button.isChecked():
            self.fast_search_button.setChecked(False)
        
        self.advanced_search_button.setEnabled(enabled)
        self.fast_search_button.setEnabled(enabled)

    def play_continously_button_on_click(self):

        # Start to play the media collection
        inst = PlayContinouslyThread.play(self)
       
        # connect the "selected" event to a method which will select the media in the drop-down list
        inst.selected.connect(self.select_play_continously_element_by_index)

    def stop_continously_button_on_click(self):
        PlayContinouslyThread.stop()

    # --------------------------
    #
    # Fast Search Button Clicked
    #
    # --------------------------
    def fast_search_button_on_click(self, checked):
        if checked:
            self.advanced_search_button.setChecked(False)
        # hide/show fast filter
        self.control_panel.fast_filter_holder.setHidden(not checked)
        # filter the list
        self.control_panel.fast_filter_on_change()
    
    # ------------------------------
    #
    # Advanced Search Button Clicked
    #
    # ------------------------------
    def advanced_search_button_on_click(self, checked):
        if checked:
            self.fast_search_button.setChecked(False)
        # hide/show advanced filter
        self.control_panel.advanced_filter_holder.setHidden(not checked)
        # filter the list
        self.control_panel.advanced_filter_filter_on_click()
        
    # -------------------
    #
    # Back Button Clicked
    #
    # -------------------
    def back_button_on_click(self):
        if self.back_button_method:
            self.back_button_method()

    # -----------------------
    #
    # Config Button Clicked
    #
    # -----------------------
    def config_button_on_click(self):

        dialog = ConfigurationDialog()

        # if OK was clicked
        if dialog.exec_() == QDialog.Accepted:        

            # get the values from the DIALOG
            l = dialog.get_language()
            mp = dialog.get_media_path()
            vp = dialog.get_media_player_video()
            vpp = dialog.get_media_player_video_param()
            ap = dialog.get_media_player_audio()
            app = dialog.get_media_player_audio_param()

            # Update the config.ini file
            config_ini_function = get_config_ini()
            config_ini_function.set_media_path(mp) 
            config_ini_function.set_language(l)
            config_ini_function.set_media_player_video(vp)
            config_ini_function.set_media_player_video_param(vpp)
            config_ini_function.set_media_player_audio(ap)
            config_ini_function.set_media_player_audio_param(app)


#!!!!!!!!!!!!
            # Re-read the config.ini file
            re_read_config_ini()

            # Re-import card_holder_pane
            mod = importlib.import_module("akoteka.gui.card_panel")
            importlib.reload(mod)
#!!!!!!!!!!!!

            # remove history
            for card_holder in self.control_panel.gui.card_holder_history:
                card_holder.setHidden(True)
                self.control_panel.gui.card_holder_panel_layout.removeWidget(card_holder)
                #self.gui.scroll_layout.removeWidget(card_holder)
            self.control_panel.gui.card_holder_history.clear()
                
            # Remove recent CardHolder as well
            self.control_panel.gui.actual_card_holder.setHidden(True)
            self.control_panel.gui.card_holder_panel_layout.removeWidget(self.control_panel.gui.actual_card_holder)
            self.control_panel.gui.actual_card_holder = None
            
            # reload the cards
            self.control_panel.gui.startCardHolder()
            
            # refresh the Control Panel
            self.control_panel.refresh_label()
            
        dialog.deleteLater()
        



class PlayContinouslyThread(QThread):
    
    selected = pyqtSignal(int)
    
    __instance = None
    __run = False
    __wait_for_stop = False
    
    @classmethod
    def play(cls, parent):
        
        if cls.__instance is None:
            cls.__new__(cls)
        
        cls.__init__(cls.__instance, parent)
            
        if not cls.__run:
            cls.__instance.start()
            
        return cls.__instance

    @classmethod
    def stop(cls):
        PlayContinouslyThread.__wait_for_stop = True
        
    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance    

    def __init__(self, parent):
        QThread.__init__(self)        
        self.parent = parent
         
    def run(self):

        PlayContinouslyThread.__run = True
        PlayContinouslyThread.__wait_for_stop = False
        self.parent.enablePlayContinously(False)

        start_index = self.parent.get_play_continously_selected_index()
        end_index = self.parent.get_play_continously_last_index()

        stop_all = False
        for index in range(start_index, end_index + 1):
            
            pid = None
            media = self.parent.get_play_continously_path_by_index(index)
            if media is not None:
            
                # play the next media
                pid = play_media(media)               
            
                # emit an media selection event
                self.selected.emit(index)                

            # checking if it stopped
            while pid is not None:

                # if the Play Continously is stopped
                if PlayContinouslyThread.__wait_for_stop:
                    
                    # then kill the process
                    stop_all = True
                    os.kill(pid, signal.SIGKILL)
                    break
                    
                    # --- Id does not work. root needed ---
                    #p = psutil.Process(pid)
                    #p.terminate()

                # I suppose the pid does not exist anymore
                still_exist_pid = False
                
                # I check the running processes
                for p in psutil.process_iter():
                    if p.pid == pid:
                        still_exist_pid = True
                        break
                    
                if not still_exist_pid:
                    break
                
                # at this point the pid is still exists so the media is playing
                # I wait a second
                QThread.sleep(1)

            # If the Play Continously is stopped
            if stop_all:
                
                # Break the loop on the files
                break

        # Indicates that the Thread is ready to start again
        self.parent.enablePlayContinously(True)
        PlayContinouslyThread.__run = False

    def __del__(self):
        self.exiting = True
        self.wait()



import sys
import time 

#from PyQt5 import Qt

from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QListWidget
from PyQt5.QtWidgets import QListWidgetItem
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QLabel

from PyQt5.QtWidgets import QCompleter
from PyQt5.QtWidgets import QDesktopWidget
from PyQt5.QtWidgets import QStyledItemDelegate
from PyQt5.QtWidgets import QStyleOptionViewItem
from PyQt5.QtWidgets import QStyle

from PyQt5.QtCore import QItemSelection
from PyQt5.QtCore import QEvent
from PyQt5.QtCore import QModelIndex
from PyQt5.QtCore import QRect
from PyQt5.QtCore import Qt
#from PyQt5.QtCore import QThread
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QSortFilterProxyModel  
from PyQt5.QtCore import QStringListModel
from PyQt5.QtCore import QAbstractListModel
from PyQt5.QtCore import QRegExp

from PyQt5 import QtCore
from PyQt5 import QtGui

from PyQt5.QtGui import QFont
from PyQt5.QtGui import QStandardItemModel, QStandardItem


#from PyQt5.QtWidgets import QAbstractItemView

# ===================
#
# AkoFilter 
#
# ===================
class AQFilter(QWidget):
    DEFAULT_MAX_LINE = 10
    DEFAULT_MIN_CHARS = 0
    
    def __init__(self, parent):
        super().__init__(parent)

        self_layout = QHBoxLayout(self)
        self_layout.setContentsMargins(0, 0, 0, 0)
        self_layout.setSpacing(0)
        self.setLayout(self_layout)

        self.parent = parent
        self.main_window = self.get_main_window()
        
        self.max_line = AQFilter.DEFAULT_MAX_LINE
        self.min_chars_to_show_list = AQFilter.DEFAULT_MIN_CHARS
        
        self.value = ""
        self.index = -1
        
        self.list = []
        
        # Input field Widget
        self.input_widget = InputLine(self)
        self.input_widget.setText(self.value)
        self_layout.addWidget(self.input_widget)
        self.input_widget.textChanged.connect(self.input_changed)
        
#        # List Widget
#        self.list_widget = QListWidget(self)
#        self.list_widget.setHidden(True)
        
        self.custom_auto_completer = CustomCompleter()
        self.custom_auto_completer.setCompletionMode(QCompleter.PopupCompletion)
        self.custom_auto_completer.setCaseSensitivity(0)
        self.input_widget.setCompleter(self.custom_auto_completer)
        self_layout.addWidget(self.input_widget)

        self.model = MyListModel(self.list)
        self.custom_auto_completer.setModel(self.model)

    def input_changed(self, value):
        self.value = value
        
    # #############
    # addItemToList
    # #############
    def addItemToList(self, value, index):
        """Add value-index pair to the  list"""
        self.list.append((value,index))
        self.custom_auto_completer.setModel(self.model)

    def setValue(self, value):
        self.value = value
        self.input_widget.setText(value)        

    # #############
    # getValue
    # #############
    def getValue(self):
        return self.value
    
    def getIndexes(self):
        return [l[1] for l in self.list if (self.value.lower() and self.value.lower() in l[0].lower())]
    
    def clear(self):
        self.value = ""
        self.index = -1
        self.input_widget.setText(self.value)
        self.list.clear()
  
    def get_main_window(self):
        """Search for the Main Window"""
        def get(widget, ret=None):
            parent = widget.parentWidget()
            if parent:
                ret = get(parent)
            else:
                ret = widget
            return ret
        
        main_window = get(self)
        return main_window


        
class InputLine(QLineEdit):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        
    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Down:
            self.parent.custom_auto_completer.complete()
        #event.ignore()
        super().keyPressEvent(event)
  
class HTMLDelegate(QStyledItemDelegate):
    """ From: https://stackoverflow.com/a/5443112/1504082 """
    def __init__(self, filter_model):
        super().__init__()
        self.filter_model = filter_model
        
    def paint(self, painter, option, index):
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        if options.widget is None:
            style = QtGui.QApplication.style()
        else:
            style = options.widget.style()

        doc = QtGui.QTextDocument()

        pattern = self.filter_model.filterRegExp().pattern()
        if pattern.lower() in options.text.lower():
            start = options.text.lower().index(pattern.lower())
            end = len(pattern) + start
            part_1 = options.text[0:start]
            part_2 = options.text[start:end]
            part_3 = options.text[end:]                
            label = part_1 + '<b>' + part_2 + '</b>' + part_3
            doc.setHtml(label)            
        else:
            doc.setHtml(options.text)

        doc.setTextWidth(option.rect.width())
        options.text = ""
        style.drawControl(QStyle.CE_ItemViewItem, options, painter)

        ctx = QtGui.QAbstractTextDocumentLayout.PaintContext()

        textRect = style.subElementRect(QStyle.SE_ItemViewItemText, options)
        painter.save()
        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        options = QStyleOptionViewItem(option)
#       self.initStyleOption(options, index)
        doc = QtGui.QTextDocument()
        doc.setHtml(options.text)
#       print(options.text)
#       doc.setTextWidth(options.rect.width())
        return QtCore.QSize(doc.size().width(), doc.size().height()-4)

        
class CustomCompleter(QCompleter):

    def __init__(self):
        super(CustomCompleter, self).__init__()
        self.source_model = None
        self.filterProxyModel = QSortFilterProxyModel(self)
        self.delegate = HTMLDelegate(self.filterProxyModel)

    def setModel(self, model):
        self.source_model = model
        self.filterProxyModel = QSortFilterProxyModel(self)
        self.filterProxyModel.setSourceModel(self.source_model)
        super(CustomCompleter, self).setModel(self.filterProxyModel)
        
        self.delegate.filter_model = self.filterProxyModel

    def updateModel(self, local_completion_prefix):
        self.filterProxyModel.setSourceModel(self.source_model)

        pattern = QRegExp(local_completion_prefix, Qt.CaseInsensitive, QRegExp.FixedString)

        self.filterProxyModel.setFilterRegExp(pattern)
                
        self.popup().setItemDelegate(self.delegate)

    def splitPath(self, path):
        self.updateModel(path)

        return []
        #if self.filterProxyModel.rowCount() == 0:
        #    self.usingOriginalModel = False
        #    self.filterProxyModel.setSourceModel(QStringListModel([path]))
        #    print(path)
        #    return [path]
        #else:
        #    return []
    

class MyListModel(QAbstractListModel):
    
    def __init__(self, mylist):
        super().__init__()
        self.mylist = mylist

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
                
        if row > len(self.mylist):
            return None
 
        if role == Qt.DisplayRole:
            return self.mylist[row][0]

        if role == Qt.EditRole:
#            self.value = self.mylist[row][0]
#            self.index = self.mylist[row][1]
            return self.mylist[row][0]
            
        return None

    def rowCount(self, parent=QModelIndex()):
        return len(self.mylist)

    def flags(self, index):
        flags = super().flags(index)
        #if index.isValid():
        #flags |= not Qt.ItemIsSelectable 
        #flags |= not Qt.ItemIsSelectable
        #   flags |= Qt.ItemIsDragEnabled
        #else:
        #    flags = Qt.ItemIsDropEnabled
        #print('flag', flags)
        flags = Qt.ItemIsEnabled
        return flags
            
#   def roleNames(self):
#       return {
#                    Qt.UserRole + 1: b'value',
#                    Qt.UserRole + 2: b'key'
#                }



class Test(QWidget):
    
    def __init__(self):
        QWidget.__init__(self)
                
        self_layout = QVBoxLayout(self)
        self_layout.setContentsMargins(10, 10, 10, 10)
        self_layout.setSpacing(0)
        self.setLayout(self_layout)
    
        # Show value field - read only
        self.show_value_field = QLineEdit()
        self.show_value_field.setReadOnly(True)
        self_layout.addWidget(self.show_value_field)
        
        # Show index field - read only
        self.show_index_field = QLineEdit()
        self.show_index_field.setReadOnly(True)
        self_layout.addWidget(self.show_index_field)

        # Show value button
        show_value_button = QPushButton("Show value")
        show_value_button.clicked.connect(self.button_clicked)
        self_layout.addWidget(show_value_button)
        
        tmp_widget = QWidget(self)
        self_layout.addWidget(tmp_widget)
        tmp_layout = QVBoxLayout(tmp_widget)        
        
        # AQFilter field
        self.ako_filter = AQFilter(tmp_widget)
#        self.ako_filter.setMinCharsToShowList(0)
        tmp_layout.addWidget(self.ako_filter)
        self.ako_filter.addItemToList("First element - plus extra text",1)
        self.ako_filter.addItemToList("Second element",2)
        self.ako_filter.addItemToList("Third element",3)
        self.ako_filter.addItemToList("Fourth element",4)
        self.ako_filter.addItemToList("Fifth element",5)
        self.ako_filter.addItemToList("Sixth element",6)
        self.ako_filter.addItemToList("Seventh element",7)
        self.ako_filter.addItemToList("Eight element",8)
        self.ako_filter.addItemToList("Nineth element",9)
        self.ako_filter.addItemToList("Tenth element",10)
        self.ako_filter.addItemToList("Eleven element",11)
        self.ako_filter.addItemToList("Twelv element",12)
        self.ako_filter.addItemToList("Thirteen element",13)
        self.ako_filter.addItemToList("Fourteen element",14)
        self.ako_filter.addItemToList("Fifteen element",15)        
      
        # --- Window ---
        self.setWindowTitle("Test AQFilter")    
        self.resize(500,200)
        self.center()
        self.show()    
        
    def button_clicked(self):
        fd = self.ako_filter
        self.show_value_field.setText(fd.getValue())
        self.show_index_field.setText(str(fd.getIndexes()))
        
    def center(self):
        """Aligns the window to middle on the screen"""
        fg=self.frameGeometry()
        cp=QDesktopWidget().availableGeometry().center()
        fg.moveCenter(cp)
        self.move(fg.topLeft())

def main():   
    
    app = QApplication(sys.argv)
    ex = Test()
    sys.exit(app.exec_())
    



import sys
import os
import importlib
#from threading import Thread
from pkg_resources import resource_string, resource_filename
from functools import cmp_to_key
import locale

from cardholder.cardholder import CardHolder
from cardholder.cardholder import Card
from cardholder.cardholder import CollectCardsThread

from akoteka.gui.pyqt_import import *
from akoteka.gui.card_panel import CardPanel
from akoteka.gui.configuration_dialog import ConfigurationDialog
from akoteka.gui.control_buttons_holder import ControlButtonsHolder

from akoteka.accessories import collect_cards
from akoteka.accessories import filter_key
from akoteka.accessories import clearLayout
from akoteka.accessories import FlowLayout


from akoteka.constants import *
from akoteka.setup.setup import getSetupIni

from akoteka.handle_property import _
from akoteka.handle_property import re_read_config_ini
from akoteka.handle_property import config_ini
from akoteka.handle_property import get_config_ini


# ==================
#
# Fast Filter HOLDER
#
# ==================
class FastFilterHolder(QWidget):
    
    changed = pyqtSignal()
    clear_clicked = pyqtSignal()
    
    def __init__(self):
        super().__init__()

        self_layout = QVBoxLayout(self)
        self.setLayout( self_layout )
        self_layout.setContentsMargins(0, 0, 0, 0)
        self_layout.setSpacing(0)    

        holder = QWidget(self)
        holder_layout = QHBoxLayout(holder)
        holder.setLayout( holder_layout )
        holder_layout.setContentsMargins(0, 0, 0, 0)
        holder_layout.setSpacing(8)    

        self_layout.addWidget(QHLine())
        self_layout.addWidget(holder)        

        # ----------
        #
        # Dropdowns 
        #
        # ----------

        #
        # Dropdown - title
        #
#        self.filter_dd_title = FilterDropDownSimple(_('title_title'))
#        holder_dropdown_gt = FilterDropDownHolder()
#      
#        holder_dropdown_gt.add_dropdown(self.filter_dd_title)
#        
#        holder_layout.addWidget(holder_dropdown_gt)        

        # -------------------------------
        #
        # Dropdown - title+director+actor
        #
        # -------------------------------
        self.filter_dd_title = FilterDropDownSimple(_('title_title') + ": ")
        self.filter_dd_director = FilterDropDownSimple(_('title_director') + ": ")
        self.filter_dd_actor = FilterDropDownSimple(_('title_actor') + ": ")
        
        holder_dropdown_da = FilterDropDownHolder()
        
        holder_dropdown_da.add_dropdown(self.filter_dd_title)
        holder_dropdown_da.add_dropdown(self.filter_dd_director)
        holder_dropdown_da.add_dropdown(self.filter_dd_actor)
        
        holder_layout.addWidget(holder_dropdown_da)

        # ----------------------
        #
        # Dropdown - genre+theme
        #
        # ----------------------
        self.filter_dd_genre = FilterDropDownSimple(_('title_genre') + ": ")
        self.filter_dd_theme = FilterDropDownSimple(_('title_theme') + ": ")
        
        holder_dropdown_gt = FilterDropDownHolder()
        
        holder_dropdown_gt.add_dropdown(self.filter_dd_genre)
        holder_dropdown_gt.add_dropdown(self.filter_dd_theme)
        
        holder_layout.addWidget(holder_dropdown_gt) 
 
        # ----------
        #
        # Checkboxes
        #
        # ----------
        self.filter_cb_favorite = FilterCheckBox(_('title_favorite') + ": ")
        self.filter_cb_best = FilterCheckBox(_('title_best') + ": ")
        self.filter_cb_new = FilterCheckBox(_('title_new') + ": ")
                
        holder_checkbox = FilterCheckBoxHolder()
        
        holder_checkbox.add_checkbox(self.filter_cb_favorite)
        holder_checkbox.add_checkbox(self.filter_cb_best)
        holder_checkbox.add_checkbox(self.filter_cb_new)        
                
        # Listener
        self.filter_cb_favorite.stateChanged.connect(self.state_changed)
        self.filter_cb_best.stateChanged.connect(self.state_changed)
        self.filter_cb_new.stateChanged.connect(self.state_changed)
                        
        holder_layout.addWidget(holder_checkbox)
 
        # ----------
        #
        # Stretch
        #
        # ----------
        holder_layout.addStretch(1)
        holder_layout.addWidget(QVLine()) 
        holder_layout.addStretch(1)
        
        # ------------
        #
        # Clear button
        #
        # ------------
        self.clear_button = QPushButton(_('button_clear'))
        holder_layout.addWidget(self.clear_button)
        self.clear_button.clicked.connect(self.clear_button_clicked)

        # Listeners
        self.filter_dd_title.state_changed.connect(self.state_changed)
        self.filter_dd_genre.state_changed.connect(self.state_changed)
        self.filter_dd_theme.state_changed.connect(self.state_changed)
        self.filter_dd_director.state_changed.connect(self.state_changed)
        self.filter_dd_actor.state_changed.connect(self.state_changed)

    def refresh_label(self):
        self.filter_dd_title.refresh_label(_('title_title'))
        self.filter_dd_genre.refresh_label(_('title_genre'))
        self.filter_dd_theme.refresh_label(_('title_theme'))
        self.filter_dd_director.refresh_label(_('title_director'))
        self.filter_dd_actor.refresh_label(_('title_actor'))
        self.filter_cb_favorite.refresh_label(_('title_favorite'))
        self.filter_cb_new.refresh_label(_('title_new'))
        self.filter_cb_best.refresh_label(_('title_best'))

    def clear_title(self):
        self.filter_dd_title.clear_elements()
        
    def add_title(self, value):
        self.filter_dd_title.add_element(value, value)

    def select_title_by_text(self, text):
        self.filter_dd_title.select_element_by_text(text)
    # ---
    def clear_genre(self):
        self.filter_dd_genre.clear_elements()
        
    def add_genre(self, value, id):
        self.filter_dd_genre.add_element(value, id)
        
    def select_genre_by_id(self, id):
        self.filter_dd_genre.select_element_by_id(id)
        
    def select_genre_by_text(self, text):
        self.filter_dd_genre.select_element_by_text(text)        
    # ---
    def clear_theme(self):
        self.filter_dd_theme.clear_elements()

    def add_theme(self, value, id):
        self.filter_dd_theme.add_element(value, id)
        
    def select_theme_by_id(self, id):
        self.filter_dd_theme.select_element_by_id(id)        

    def select_theme_by_text(self, text):
        self.filter_dd_theme.select_element_by_text(text)        

    # ---
    def clear_director(self):
        self.filter_dd_director.clear_elements()
    
    def add_director(self, director):
        self.filter_dd_director.add_element(director, director)
    
    def select_director_by_text(self, text):
        self.filter_dd_director.select_element_by_text(text)
    # ---
    def clear_actor(self):
        self.filter_dd_actor.clear_elements()

    def add_actor(self, actor):
        self.filter_dd_actor.add_element(actor, actor)
        
    def select_actor_by_text(self, text):
        self.filter_dd_actor.select_element_by_text(text)        
    # ---
    def get_filter_selection(self):
        filter_selection = {
            "title":    ["title", self.filter_dd_title.get_selected_value(), None, "a"],
            "genre":    ["genre", self.filter_dd_genre.get_selected_value(), [self.filter_dd_genre.get_selected_id()], "a"],
            "theme":    ["theme", self.filter_dd_theme.get_selected_value(), [self.filter_dd_theme.get_selected_id()], "a"],
            "director": ["director", self.filter_dd_director.get_selected_value(), None, "a"],
            "actor":    ["actor", self.filter_dd_actor.get_selected_value(), None, "a"],
            "best":     ["best", self.filter_cb_best.is_checked(), None, "c"],
            "new":      ["new", self.filter_cb_new.is_checked(), None, "c"],
            "favorite": ["favorite", self.filter_cb_favorite.is_checked(),None, "c"],
        }
        return filter_selection
    
    def clear_button_clicked(self):
        self.clear_clicked.emit()
    
    def state_changed(self):
        self.changed.emit()

    def closeEvent(self, event):
        print('close filter holder')
        
    def clear_fields(self):
        self.clear_actor()
        self.clear_director()
        self.clear_genre()
        self.clear_theme()
        self.clear_title()
        self.filter_cb_favorite.setChecked(False)
        self.filter_cb_new.setChecked(False)
        self.filter_cb_best.setChecked(False)


# ================
#
# Dropdown HOLDER
#
# ================
class FilterDropDownHolder(QWidget):
    
    def __init__(self):
        super().__init__()

        self.self_layout = QVBoxLayout(self)
        self.setLayout( self.self_layout )
        self.self_layout.setContentsMargins(0, 0, 0, 0)
        self.self_layout.setSpacing(1)
        self.self_layout.setAlignment(Qt.AlignTop)

#        self.setStyleSheet( 'background: green')

    def add_dropdown(self, filter_dropdown):
        self.self_layout.addWidget(filter_dropdown)

# =============================
#
# Filter Drop-Down Simple
#
# =============================
#
class FilterDropDownSimple(QWidget):
    
    state_changed = pyqtSignal()
    
    def __init__(self, label):
        super().__init__()

        self_layout = QHBoxLayout(self)
        self_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout( self_layout )
#        self.setStyleSheet( 'background: green')
        
        self.label_widget = QLabel(label)
        self_layout.addWidget( self.label_widget )

        self.dropdown = QComboBox(self)
        self.dropdown.setFocusPolicy(Qt.NoFocus)
        #self.dropdown.setEditable(True)
        
        self.dropdown.currentIndexChanged.connect(self.current_index_changed)

        # TODO does not work to set the properties of the dropdown list. find out and fix
        style =             '''
            QComboBox { max-width: 200px; min-width: 200px; min-height: 15px; max-height: 15px;}
            QComboBox QAbstractItemView::item { min-width:100px; max-width:100px; min-height: 150px;}
            QListView::item:selected { color: red; background-color: lightgray; min-width: 1000px;}"
            '''            

        style_down_arrow = '''
            QComboBox::down-arrow { 
                image: url( ''' + resource_filename(__name__,os.path.join("img", "back-button.jpg")) + ''');
                
            }
        '''
        style_box = '''
            QComboBox { 
                max-width: 200px; min-width: 200px; border: 1px solid gray; border-radius: 5px;
            }
        '''
#       max-width: 200px; min-width: 200px; min-height: 1em; max-height: 1em; border: 1px solid gray; border-radius: 5px;
        
        style_drop_down ='''
            QComboBox QAbstractItemView::item { 
                color: red;
                max-height: 15px;
            }
        '''            
      
        self.dropdown.setStyleSheet(style_box + style_drop_down)
        self.dropdown.addItem("")

        self_layout.addWidget( self.dropdown )
    
    def clear_elements(self):
        self.dropdown.clear()

    def add_element(self, value, id):
        self.dropdown.addItem(value, id)

    # -------------------------------------
    # get the index of the selected element
    # -------------------------------------
    def get_selected_id(self):
        return self.dropdown.itemData( self.dropdown.currentIndex() )

    # -------------------------------------
    # get the value of the selected element
    # -------------------------------------
    def get_selected_value(self):
        return self.dropdown.itemText( self.dropdown.currentIndex() )
    
    def select_element_by_id(self, id):
        self.dropdown.setCurrentIndex( self.dropdown.findData(id) )

    def select_element_by_text(self, text):
        self.dropdown.setCurrentIndex( self.dropdown.findText(text) )

    def current_index_changed(self):
        self.state_changed.emit()
        
    def refresh_label(self, new_label):
        self.label_widget.setText(new_label)


# ==========
#
# CheckBox
#
# ==========
class FilterCheckBox(QCheckBox):
    def __init__(self, label):
        super().__init__(label)

        self.setLayoutDirection( Qt.RightToLeft )
        style_checkbox = '''
            QCheckBox { 
                min-height: 15px; max-height: 15px; border: 0px solid gray;
            }
        '''
        self.setStyleSheet( style_checkbox )
#        self.setFocusPolicy(Qt.NoFocus)

    def is_checked(self):
        return 'y' if self.isChecked() else None        
 
    def refresh_label(self, new_label):
        self.setText(new_label)


# ================
#
# Checkbox HOLDER
#
# ================
class FilterCheckBoxHolder(QWidget):
    
    def __init__(self):
        super().__init__()

        self.self_layout = QVBoxLayout(self)
        self.setLayout( self.self_layout )
        self.self_layout.setContentsMargins(0, 0, 0, 0)
        self.self_layout.setSpacing(1)

        #self.setStyleSheet( 'background: green')
        
    def add_checkbox(self, filter_checkbox):
        self.self_layout.addWidget(filter_checkbox)
        







import sys
import os
import importlib
from threading import Thread
from pkg_resources import resource_string, resource_filename
from functools import cmp_to_key
import locale

from cardholder.cardholder import CardHolder
from cardholder.cardholder import Card
from cardholder.cardholder import CollectCardsThread

from akoteka.gui.pyqt_import import *
from akoteka.gui.card_panel import CardPanel
from akoteka.gui.configuration_dialog import ConfigurationDialog
from akoteka.gui.control_buttons_holder import ControlButtonsHolder
from akoteka.gui.fast_filter_holder import FastFilterHolder
from akoteka.gui.advanced_filter_holder import AdvancedFilterHolder

from akoteka.accessories import collect_cards
from akoteka.accessories import filter_key
from akoteka.accessories import clearLayout
from akoteka.accessories import FlowLayout

from akoteka.constants import *
from akoteka.setup.setup import getSetupIni

from akoteka.handle_property import _
from akoteka.handle_property import re_read_config_ini
from akoteka.handle_property import config_ini
from akoteka.handle_property import get_config_ini
#from builtins import None

class GuiAkoTeka(QWidget, QObject):
    
    def __init__(self):
        #super().__init__()  
        QWidget.__init__(self)
        QObject.__init__(self)
        
        self.actual_card_holder = None
        self.card_holder_history = []
        
        # most outer container, just right in the Main Window
        box_layout = QVBoxLayout(self)
        self.setLayout(box_layout)
        # controls the distance between the MainWindow and the added container: scrollContent
        box_layout.setContentsMargins(0, 0, 0, 0)
        box_layout.setSpacing(0)
    
        # control panel
        self.control_panel = ControlPanel(self)
        self.control_panel.set_back_button_method(self.restore_previous_holder)
        box_layout.addWidget( self.control_panel)
    
        # scroll_content where you can add your widgets - has scroll
        scroll = QScrollArea(self)
        box_layout.addWidget(scroll)
        scroll.setWidgetResizable(True)
        scroll_content = QWidget(scroll)
        scroll_content.setStyleSheet('background: ' + COLOR_MAIN_BACKGROUND)
        scroll.setFocusPolicy(Qt.NoFocus)
    
#        scroll_content = QWidget(self)
#        scroll_content.setStyleSheet('background: ' + COLOR_MAIN_BACKGROUND)
#        box_layout.addWidget(scroll_content)
#        scroll_layout = QVBoxLayout(scroll_content)
#        scroll_content.setLayout(scroll_layout)

        # layout of the content with margins
        scroll_layout = QVBoxLayout(scroll_content)        
        scroll.setWidget(scroll_content)        
        # vertical distance between cards - Vertical
        scroll_layout.setSpacing(5)
        # spaces between the added Widget and this top, right, bottom, left side
        scroll_layout.setContentsMargins(15,15,15,15)
        scroll_content.setLayout(scroll_layout)

        # -------------------------------
        # Title
        # -------------------------------
        self.hierarchy_title = HierarchyTitle(scroll_content, self)
        self.hierarchy_title.set_background_color(QColor(COLOR_CARDHOLDER_BACKGROUND))
        self.hierarchy_title.set_border_radius(RADIUS_CARDHOLDER)

        # -------------------------------
        # Here comes later the CardHolder
        # -------------------------------
        self.card_holder_panel = QWidget(scroll_content)
        
        scroll_layout.addWidget(self.hierarchy_title)
        scroll_layout.addWidget(self.card_holder_panel)
        scroll_layout.addStretch(1)
        
        self.card_holder_panel_layout = QVBoxLayout(self.card_holder_panel)
        self.card_holder_panel_layout.setContentsMargins(0,0,0,0)
        self.card_holder_panel_layout.setSpacing(0)

        self.back_button_listener = None

        # --- Window ---
        sp=getSetupIni()
        self.setWindowTitle(sp['name'] + '-' + sp['version'])
        self.setWindowIcon(QIcon(resource_filename(__name__,os.path.join("img", IMG_WINDOW)))) 
        #self.setGeometry(300, 300, 300, 200)
        self.resize(900,600)
        self.center()
        self.show()    

    def center(self):
        """Aligns the window to middle on the screen"""
        fg=self.frameGeometry()
        cp=QDesktopWidget().availableGeometry().center()
        fg.moveCenter(cp)
        self.move(fg.topLeft())

    def startCardHolder(self):
        """
        --- Start Card Holder ---
        
        """
        # Create the first Card Holder - 
        #self.go_down_in_hierarchy( [], "" )
        self.go_down_in_hierarchy() 

        # Retreive the media path
        paths = [config_ini['media_path']]
        
        # Start to collect the Cards from the media path
        self.actual_card_holder.startCardCollection(paths)        

    def go_down_in_hierarchy( self, card_descriptor_structure=None, title=None, save=True ):
        """
        --- Go Down in Hierarchy
        
        Go deeper in the hierarchy

         card_descriptor_structure The NOT Filtered card hierarchy list
                                   on the recent level
         title                     The title what whould be shown above
                               the CardHolder
         save                      It controls to save the CardHolder
                                   into the history list
                                   collecting_finish uses it with False
        """
        # if it is called very first time
        if card_descriptor_structure is None:
            self.initialize = True
            card_descriptor_structure = []
            title = ""
            save = False
        else:
            self.initialize = False

        # if there is already a CardHolder
        if self.actual_card_holder:

            # hide the old CardHolder
            self.actual_card_holder.setHidden(True)

        # if it is said to Save the CardHolder to the history list
        if save:
            
            # save the old CardHolder it in a list
            self.card_holder_history.append(self.actual_card_holder)
                    
        self.actual_card_holder = CardHolder(            
            self, 
            self.get_new_card,
            self.collect_cards,
            self.collecting_start,
            self.collecting_finish
        )
        
        self.actual_card_holder.title = title
        self.actual_card_holder.set_max_overlapped_cards( MAX_OVERLAPPED_CARDS )
        self.actual_card_holder.set_y_coordinate_by_reverse_index_method(self.get_y_coordinate_by_reverse_index)
        self.actual_card_holder.set_x_offset_by_index_method(self.get_x_offset_by_index)
        self.actual_card_holder.set_background_color(QColor(COLOR_CARDHOLDER_BACKGROUND))
        self.actual_card_holder.set_border_radius(RADIUS_CARDHOLDER)
        self.actual_card_holder.set_border_width(15)        
                
        # Save the original card desctiptor structure into the CardHolder
        self.actual_card_holder.orig_card_descriptor_structure = card_descriptor_structure
        
        # Make the CardHolder to be in Focus
        self.actual_card_holder.setFocus()

        # Set the title of the CardHolder - The actual level of the hierarchy
        self.hierarchy_title.set_title(self.card_holder_history, self.actual_card_holder)

        # add the new holder to the panel
        self.card_holder_panel_layout.addWidget(self.actual_card_holder)
        #self.scroll_layout.addStretch(1)

        # filter the list by the filters + Fill-up the CardHolder with Cards using the parameter as list of descriptor
        self.filter_the_cards(card_descriptor_structure)

    def restore_previous_holder(self, steps=1):
        """
        --- Restore Previous Holder ---
        
        Come up in the hierarchy
        """        
        size = len(self.card_holder_history)
        if  size >= steps:

            for i in range(0, steps):
            
                previous_card_holder = self.card_holder_history.pop()
                # get the previous CardHolder
                #previous_card_holder = self.card_holder_history[size - 1]
            
                # remove the previous CardHolder from the history list
                #self.card_holder_history.remove(previous_card_holder)
            
            # hide the old CardHolder
            self.actual_card_holder.setHidden(True)            
            
            # remove from the layout the old CardHolder
            self.card_holder_panel_layout.removeWidget(self.actual_card_holder)
        
            # the current card holder is the previous
            self.actual_card_holder = previous_card_holder
            
            # show the current card holder
            self.actual_card_holder.setHidden(False)

            # set the title
            self.hierarchy_title.set_title(self.card_holder_history, self.actual_card_holder)
            
            # filter the list by the filters + Fill-up the CardHolder with Cards using the parameter as list of descriptor
            self.filter_the_cards(self.actual_card_holder.orig_card_descriptor_structure)
            #self.filter_the_cards(self.actual_card_holder.orig_card_descriptor_structure)
            
            # select the Card which was selected to enter
            self.actual_card_holder.select_actual_card()
            
            # Make the CardHolder to be in Focus
            self.actual_card_holder.setFocus()         

    def collecting_start(self):
        """
        --- Collecting Started ---
        Indicates that the CardCollection process started.
        The CardHolder calls this method
        Jobs to do:
            -Hide the title
        """
        self.hierarchy_title.setHidden(True)
        
        # close the Search panels and disable buttons to search
        self.control_panel.control_buttons_holder.enableSearchIcons(False)
        self.control_panel.control_buttons_holder.disablePlayStopContinously()

    def collecting_finish(self, card_holder, card_descriptor_structure):
        """
        --- Collecting Finished ---
        
        Indicates that the CardCollection process finished.
        The CardHolder calls this method
        Jobs to do:
            -Show the title
            -Set up the filters
        """        
        
        if card_descriptor_structure:
          
            # Show the title of the CardHolder (the certain level)        
            self.hierarchy_title.setHidden(False)       
        
        # Save the NOT Filtered list
        card_holder.orig_card_descriptor_structure = card_descriptor_structure
        
        # Set-up the Filters        
        self.set_up_filters(card_descriptor_structure)
        
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # This part is tricky
        # It prevents to show the 0. level of Cards
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        #
        # if there is only ONE Card and the status of the presentation is "initialize"
        if len(card_holder.card_descriptor_list) == 1 and self.initialize:
            # then I go down ONE level
#            self.go_down_in_hierarchy(card_holder.card_descriptor_list[0]['extra']["orig-sub-cards"], card_holder.card_descriptor_list[0]['title'][config_ini['language']], save=False )
            self.go_down_in_hierarchy(card_holder.card_descriptor_list[0]['extra']["sub-cards"], card_holder.card_descriptor_list[0]['title'][config_ini['language']], save=False )

        # Enable the buttons to search
        self.control_panel.control_buttons_holder.enableSearchIcons(True)
        self.control_panel.control_buttons_holder.enablePlayContinously(True)
        
    def get_y_coordinate_by_reverse_index(self, reverse_index, diff_to_max):
        """
        --- Get Y coordinate by reverse index
      
        Calculates the Y coordinate of a Card using reverse index        
        """
        raw_pos = (reverse_index + diff_to_max) * (reverse_index + diff_to_max)
        offset = diff_to_max * diff_to_max
        return (raw_pos - offset) * 8
        #return reverse_index * 220
    
    def get_x_offset_by_index(self, index):
        """
        --- Get X offset by index ---
        
        Calculates the X coordinate of a Card by index        
        """
        return index * 4       

    def collect_cards(self, paths):
        """
        --- Collect Cards ---
        
        It will be executed only once
        in the CardHolder.CollectCardThread.run() what the 
        CardHOlder.startCardCollection() triggers in the startHolder() method
        """
        cdl = collect_cards(paths)
        filtered_card_structure = self.get_cards_by_advanced_filter(cdl)
        return filtered_card_structure
    
    def get_new_card(self, card_data, local_index, index):
        """
        --- Get New Card ---
        
        Generates a new Card
        """

        card = Card(self.actual_card_holder, card_data, local_index, index)
        
        if card_data["extra"]["media-path"]:
            card.set_background_color(QColor(COLOR_CARD_MOVIE_BACKGROUND))
        else:
            card.set_background_color(QColor(COLOR_CARD_COLLECTOR_BACKGROUND))
            
        card.set_border_normal_color(QColor(COLOR_CARD_BORDER_NORMAL_BACKGROUND))
        card.set_border_selected_color(QColor(COLOR_CARD_BORDER_SELECTED_BACKGROUND))
        
        card.set_not_selected()
        card.set_border_radius( RADIUS_CARD )
        card.set_border_width( WIDTH_BORDER_CARD )
 
        panel = card.get_panel()        
        layout = panel.get_layout()
        layout.setContentsMargins(0, 0, 0, 0)

        card_panel = CardPanel(card, card_data)
        layout.addWidget(card_panel)
        
        return card
  
    def set_fast_filter_listener(self, listener):
        self.control_panel.set_fast_filter_listener(listener)
        
    def set_advanced_filter_listener(self, listener):
        self.control_panel.set_advanced_filter_listener(listener)
        
    def get_fast_filter_holder(self):
        return self.control_panel.get_fast_filter_holder()
    
    def get_advanced_filter_holder(self):
        return self.control_panel.get_advanced_filter_holder()
      
    # --------------------------------
    #
    # Filter the Cards
    #
    # Filters the Cards and Show them
    #
    # --------------------------------
    def filter_the_cards(self, card_descriptor_structure=None):
        if card_descriptor_structure is None:
            card_descriptor_structure = self.actual_card_holder.orig_card_descriptor_structure
        
        filtered_card_structure = self.set_up_filters(card_descriptor_structure)
        
        # Fill up the CardHolder with Cards, alpabetically orderd by title 
        self.actual_card_holder.fillUpCardHolderByDescriptor(sorted(filtered_card_structure, key=lambda arg: locale.strxfrm(arg['title'][config_ini['language']]), reverse=False))
            
            
    # ----------------
    # Set-up Filters
    # ----------------
    def set_up_filters(self, card_descriptor_structure):
        # --- FILL UP ADVANCED FILTERS ---
        unconditional_list = self.get_unconditional_filter_list(card_descriptor_structure)        
        self.fill_up_advanced_filter_lists(unconditional_list)
        
        # --- FILL UP FAST FILTERS ---        
        fast_filter_list = self.get_fast_filter_list(card_descriptor_structure)
        self.fill_up_fast_filter_lists(fast_filter_list)

        # --- Filter the cards setting its ['extra']['visible'] attribute ---
        if not self.control_panel.fast_filter_holder.isHidden():
            filtered_cards = self.get_cards_by_fast_filter(card_descriptor_structure)
        else:    
            filtered_cards = self.get_cards_by_advanced_filter(card_descriptor_structure)
        
        # --- FILL UP PLAY CONTINOUSLY DROP-DOWN ---
        self.fill_up_play_continously_list(filtered_cards)
        
        return filtered_cards
    
        
    
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
    def get_fast_filter_list(self, card_structure):
        """ --------------------------------------------------------------------------------
        
        --- Get Fast Filter List ---
        
        Gives back the filter lists in the certain point of the card hierarchy                
        ------------------------------------------------------------------------------------""" 
        
        def generate_fast_filter_list(card_structure):
        
            # through the SORTED list
            #for crd in sorted(card_structure, key=lambda arg: arg['title'][config_ini['language']], reverse=False):
            for crd in card_structure:

                # in case of MEDIA CARD
                if crd['extra']['media-path']:              

                    # if the card fits to the filter
                    if self.isFitByFilterSelection(crd, self.get_fast_filter_holder().get_filter_selection().items()):

                        for c, v in self.get_fast_filter_holder().get_filter_selection().items():
                            category = v[0]
                        
                            if filter_key[category]['store-mode'] == 'v':
                                if crd[filter_key[category]['section']][category]:
                                    fast_filter_list[category].add(crd[filter_key[category]['section']][category])
                                
                            elif filter_key[category]['store-mode'] == 'a':
                                for cat in crd[filter_key[category]['section']][category]:
                                    if cat.strip():
                                        fast_filter_list[category].add(cat.strip())
                
                            elif filter_key[category]['store-mode'] == 't':
                                fast_filter_list['title'].add(crd[filter_key[category]['section']][config_ini['language']])
                    
                # in case of COLLECTOR CARD
                else:                     

                    # then it depends on the next level
                    generate_fast_filter_list(crd['extra']['sub-cards'])
        
            return fast_filter_list  
  
        fast_filter_list={
            "title": set(),
            "genre": set(),
            "theme": set(),
            "director": set(),
            "actor": set(),
#            "sound": set(),
#            "sub": set(),
#            "country": set(),
#            "year": set(),
#            "length": set(),
            "favorite": set(),
            "new": set(),
            "best": set() 
        }

        generate_fast_filter_list(card_structure)
        
        return fast_filter_list
    
    def fill_up_fast_filter_lists(self, fast_filter_list):
        """--------------------------------------------------------------
        
        Fill up the Fast Filter list by the fast_filter_list
        
        -----------------------------------------------------------------"""
        self.set_fast_filter_listener(None)
        
        fast_state_fields = {
                "title": "",
                "genre": "",
                "theme": "",
                "director": "",
                "actor": "",
#                "sound": "",
#                "sub": "",
#                "country": "",
#                "length-from": "",
#                "length-to": "",
#                "year-from": "",
#                "year-to": "",            
            }        

        # save the fields of the filters because the will be clear when the list is updated
        for category, value in self.get_fast_filter_holder().get_filter_selection().items():            
            if value[1] != None and value[1] != "":
                fast_state_fields[category] = value[1]

        # Fill up TITLE 
        self.get_fast_filter_holder().clear_title()
        self.get_fast_filter_holder().add_title("")
        for element in sorted( fast_filter_list['title'], key=cmp_to_key(locale.strcoll) ):
            self.get_fast_filter_holder().add_title(element)

        # Fill up GENRE 
        self.get_fast_filter_holder().clear_genre()
        self.get_fast_filter_holder().add_genre("", "")
        for element in sorted([(_("genre_" + e),e) for e in fast_filter_list['genre']], key=lambda t: locale.strxfrm(t[0]) ):            
            self.get_fast_filter_holder().add_genre(element[0], element[1])
        
        # Fill up THEME 
        self.get_fast_filter_holder().clear_theme()
        self.get_fast_filter_holder().add_theme("", "")
        for element in sorted([(_("theme_" + e), e) for e in fast_filter_list['theme']], key=lambda t: locale.strxfrm(t[0]) ):
            self.get_fast_filter_holder().add_theme(element[0], element[1])

        # Fill up DIRECTOR 
        self.get_fast_filter_holder().clear_director()
        self.get_fast_filter_holder().add_director("")
        for element in sorted( fast_filter_list['director'], key=cmp_to_key(locale.strcoll) ):
            self.get_fast_filter_holder().add_director(element)

        # Fill up ACTOR 
        self.get_fast_filter_holder().clear_actor()
        self.get_fast_filter_holder().add_actor("")
        for element in sorted( fast_filter_list['actor'], key=cmp_to_key(locale.strcoll) ):
            self.get_fast_filter_holder().add_actor(element)

        # Recover the contents of the fields
        self.get_fast_filter_holder().select_title_by_text(fast_state_fields["title"])
        self.get_fast_filter_holder().select_genre_by_text(fast_state_fields["genre"])
        self.get_fast_filter_holder().select_theme_by_text(fast_state_fields["theme"])
        self.get_fast_filter_holder().select_director_by_text(fast_state_fields["director"])
        self.get_fast_filter_holder().select_actor_by_text(fast_state_fields["actor"])        

        self.set_fast_filter_listener(self)
        
    def get_cards_by_fast_filter(self, card_structure):
        """
        parameters:
            card_structure:        the full structure on the recent level (not filtered)
        """  
        def generate_filtered_list(card_structure):

            mediaFits = False
            collectorFits = False
            
            # through the card structure
            for crd in card_structure:
            
                # in case of MEDIA CARD
                if crd['extra']['media-path']:

                    fits = True

                    if not self.control_panel.fast_filter_holder.isHidden():
    
                        fits = self.isFitByFilterSelection(crd, self.get_fast_filter_holder().get_filter_selection().items())
                 
                    if fits:

                        mediaFits = True
                 
                    crd['extra']['visible'] = fits
   
                # in case of COLLECTOR CARD
                else:                     

                    # then it depends on the next level
                    fits = generate_filtered_list(crd['extra']['sub-cards'])
                
                    if fits:
                        collectorFits = True        
            
                    crd['extra']['visible'] = fits
                
            return (mediaFits or collectorFits)
    
        generate_filtered_list(card_structure)
        return card_structure    
    







    
    def fill_up_play_continously_list(self, filtered_card_structure):
        """
        Gives back a list of media-cards which should be played
        
        Parameters:
            filtered_card_structure   ['extra']['sub-cards'] False - should not be played (filtered out) 
        
        Return:
            <list>[0] True/False - media/separator
            <list>[1] path to media
            <list>[2] title of the media
        """

        def generate_list(filtered_card_structure, playing_list):

            # through the SORTED list
            #for crd in sorted(filtered_card_structure, key=lambda arg: locale.strxfrm(arg['title'][config_ini['language']]), reverse=False):
            for crd in sorted(filtered_card_structure, key=lambda arg:  locale.strxfrm(arg['title'][config_ini['language']])    if arg['extra']['media-path'] and arg['extra']['visible'] else "_" + locale.strxfrm(arg['title'][config_ini['language']])            , reverse=False):
            
                # in case of MEDIA CARD
                if crd['extra']['media-path'] and crd['extra']['visible']:

                    playing_list.append((True, crd['extra']['media-path'], crd['title'][config_ini['language']]))
   
                # in case of COLLECTOR CARD
                elif crd['extra']['visible']:                     

                    playing_list.append((False, crd['extra']['media-path'], crd['title'][config_ini['language']]))
                    
                    # then it depends on the next level
                    generate_list(crd['extra']['sub-cards'], playing_list)
                
            return
    
        playing_list = []        
        generate_list(filtered_card_structure, playing_list)
        
        self.control_panel.control_buttons_holder.clear_play_continously_elements()
        self.control_panel.control_buttons_holder.add_play_continously_element("", "")
        for l in playing_list:
            if l[0]:
                self.control_panel.control_buttons_holder.add_play_continously_element(l[2], l[1])
            else:
                self.control_panel.control_buttons_holder.add_play_continously_separator()

    
    
    
    
    
    
    
      
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
    def get_unconditional_filter_list(self, card_structure):
        """ --------------------------------------------------------------------------------
        
        --- Get Advanced Filter List ---
        
        Gives back the unconditional filter lists in the certain point of the card hierarchy                
        ------------------------------------------------------------------------------------""" 
        
        def generate_unconditional_filter_list(card_structure):
        
            # through the SORTED list
            for crd in card_structure:
            #for crd in sorted(card_structure, key=lambda arg: arg['title'][config_ini['language']], reverse=False):

                # in case of MEDIA CARD
                if crd['extra']['media-path']:              

                    for c, v in self.get_advanced_filter_holder().get_filter_selection().items():
                        category = v[0]
                        
                        if filter_key[category]['store-mode'] == 'v':
                            if crd[filter_key[category]['section']][category]:
                                unconditional_filter_list[category].add(crd[filter_key[category]['section']][category])
                                
                        elif filter_key[category]['store-mode'] == 'a':
                            for cat in crd[filter_key[category]['section']][category]:
                                if cat.strip():
                                    unconditional_filter_list[category].add(cat.strip())
                
                        elif filter_key[category]['store-mode'] == 't':
                            unconditional_filter_list['title'].add(crd[filter_key[category]['section']][config_ini['language']])
                    
                # in case of COLLECTOR CARD
                else:                     

                    # then it depends on the next level
                    generate_unconditional_filter_list(crd['extra']['sub-cards'])
        
            return unconditional_filter_list  
  
        unconditional_filter_list={
            "title": set(),
            "genre": set(),
            "theme": set(),
            "director": set(),
            "actor": set(),
            "sound": set(),
            "sub": set(),
            "country": set(),
            "year": set(),
            "length": set(),
            "favorite": set(),
            "new": set(),
            "best": set() }

        # ADVANCED FILTER is visible
        #if not self.control_panel.advanced_filter_holder.isHidden():
        generate_unconditional_filter_list(card_structure)
        
        return unconditional_filter_list
  
    def fill_up_advanced_filter_lists(self, unconditional_filter_list):
        """--------------------------------------------------------------
        
        Fill up the Advanced Filter list by the unconditional_filter_list
        
        -----------------------------------------------------------------"""

        advanced_state_fields = {
                "title": "",
                "genre": "",
                "theme": "",
                "director": "",
                "actor": "",
                "sound": "",
                "sub": "",
                "country": "",
                "length-from": "",
                "length-to": "",
                "year-from": "",
                "year-to": "",            
            }        

        # save the fields of the filters because the will be clear when the list is updated
        for category, value in self.get_advanced_filter_holder().get_filter_selection().items():
            if value[1] != None and value[1] != "":
                advanced_state_fields[category] = value[1]

        # Fill up TITLE
        self.get_advanced_filter_holder().clear_title()            
        for element in sorted( unconditional_filter_list['title'], key=cmp_to_key(locale.strcoll) ):
            self.get_advanced_filter_holder().add_title(element)
        
        # Fill up GENRE 
        self.get_advanced_filter_holder().clear_genre()            
        for element in sorted([(_("genre_" + e),e) for e in unconditional_filter_list['genre']], key=lambda t: locale.strxfrm(t[0]) ):            
            self.get_advanced_filter_holder().add_genre(element[0], element[1])
        
        # Fill up THEME 
        self.get_advanced_filter_holder().clear_theme()
        for element in sorted([(_("theme_" + e), e) for e in unconditional_filter_list['theme']], key=lambda t: locale.strxfrm(t[0]) ):            
            self.get_advanced_filter_holder().add_theme(element[0], element[1])

        # Fill up DIRECTOR 
        self.get_advanced_filter_holder().clear_director()            
        for element in sorted( unconditional_filter_list['director'], key=cmp_to_key(locale.strcoll) ):
            self.get_advanced_filter_holder().add_director(element)

        # Fill up ACTOR 
        self.get_advanced_filter_holder().clear_actor()            
        for element in sorted( unconditional_filter_list['actor'], key=cmp_to_key(locale.strcoll) ):
            self.get_advanced_filter_holder().add_actor(element)
        
        # Fill up SOUND
        self.get_advanced_filter_holder().clear_sound()
        #self.get_advanced_filter_holder().add_sound("", "")
        for element in sorted([(_("lang_long_" + e), e) for e in unconditional_filter_list['sound']], key=lambda t: locale.strxfrm(t[0]) ):            
            self.get_advanced_filter_holder().add_sound(element[0], element[1])

        # Fill up SUBTITLE
        self.get_advanced_filter_holder().clear_sub()
        #self.get_advanced_filter_holder().add_sub("", "")
        for element in sorted([(_("lang_long_" + e), e) for e in unconditional_filter_list['sub']], key=lambda t: locale.strxfrm(t[0]) ):            
            self.get_advanced_filter_holder().add_sub(element[0], element[1])

        # Fill up COUNTRY
        self.get_advanced_filter_holder().clear_country()
        #self.get_advanced_filter_holder().add_country("", "")
        for element in sorted([(_("country_long_" + e), e) for e in unconditional_filter_list['country']], key=lambda t: locale.strxfrm(t[0]) ):            
            self.get_advanced_filter_holder().add_country(element[0], element[1])
       
        # Fill up YEAR from/to
        self.get_advanced_filter_holder().clear_year()
        #self.get_advanced_filter_holder().add_year("")
        for element in sorted( unconditional_filter_list['year'], key=cmp_to_key(locale.strcoll) ):
            self.get_advanced_filter_holder().add_year(element)

        # Fill up LENGTH from/to
        self.get_advanced_filter_holder().clear_length()
        #self.get_advanced_filter_holder().add_length("")
        for element in sorted( [str(int(spl[-2])).rjust(1) + ":" + str(int(spl[-1])).zfill(2) for spl in [l.split(":") for l in unconditional_filter_list['length'] if ':' in l ] if all(c.isdigit() for c in spl[-1] ) and all(c.isdigit() for c in spl[-2] )], key=cmp_to_key(locale.strcoll) ):
            self.get_advanced_filter_holder().add_length(element)
  
        # Recover the contents of the fields
        self.get_advanced_filter_holder().set_title(advanced_state_fields['title'])
        self.get_advanced_filter_holder().set_genre(advanced_state_fields['genre'])
        self.get_advanced_filter_holder().set_theme(advanced_state_fields['theme'])
        self.get_advanced_filter_holder().set_director(advanced_state_fields['director'])
        self.get_advanced_filter_holder().set_actor(advanced_state_fields['actor'])
        self.get_advanced_filter_holder().select_sound(advanced_state_fields['sound'])
        self.get_advanced_filter_holder().select_sub(advanced_state_fields['sub'])
        self.get_advanced_filter_holder().select_country(advanced_state_fields['country'])
        self.get_advanced_filter_holder().select_year_from(advanced_state_fields['year-from'])
        self.get_advanced_filter_holder().select_year_to(advanced_state_fields['year-to'])
        self.get_advanced_filter_holder().select_length_from(advanced_state_fields['length-from'])
        self.get_advanced_filter_holder().select_length_to(advanced_state_fields['length-to'])
        
    def get_cards_by_advanced_filter(self, card_structure):
        """
        parameters:
            card_structure:        the full structure on the recent level (not filtered)
        """  
        def generate_filtered_list(card_structure):

            mediaFits = False
            collectorFits = False
            
            # through the card structure
            for crd in card_structure:
            
                # in case of MEDIA CARD
                if crd['extra']['media-path']:

                    fits = True

                    if not self.control_panel.advanced_filter_holder.isHidden():
    
                        fits = self.isFitByFilterSelection(crd, self.get_advanced_filter_holder().get_filter_selection().items())
                 
                    if fits:

                        mediaFits = True

                    crd['extra']['visible'] = fits
   
                # in case of COLLECTOR CARD
                else:                     

                    # then it depends on the next level
                    fits = generate_filtered_list(crd['extra']['sub-cards'])
                
                    if fits:
                        collectorFits = True        
            
                    crd['extra']['visible'] = fits
                
            return (mediaFits or collectorFits)
    
        generate_filtered_list(card_structure)
        return card_structure

    def to_integer(self, value):         
        hours, minutes = map(int, (['0']+value.split(':'))[-2:])
        return hours * 60 + minutes
  
    def isFitByFilterSelection(self, card, filter_selection):
        """
        
        --- Is Fit By Filter Selection ---
        
        Say if the card fits to the Filter Selection        
        """
        fits = True            
        # go through the ADVANCED FILTERS by Categories and decide if the Card is filtered
        for c, v in filter_selection:
            category = v[0]
                        
            # do I want to check this Category match
            if v[1]:                            
                            
                fits = False
                operation = v[3]
                                
                if filter_key[category]['value-dict']:
                    mylist = v[2]
                else:
                    mylist = [v[1]]                                 
                
                for filter in mylist:
                    fits = False
                                    
                    # if multiple category values 
                    if filter_key[category]['store-mode'] == 'a':
                                      
                        # go through the category values in the card
                        # at least one category value should match to the filter
                        for e in card[filter_key[category]['section']][category]:

                            # is the filter a DICT
                            if filter_key[category]['value-dict']:

                                # then correct match needed
                                if filter.lower() == e.lower():
                                    fits = True
                                    break
                                        
                            # NOT dict
                            else:
                                                
                                #NOT correct mach needed
                                if filter.lower() in e.lower():
                                    fits = True
                                    break

                    elif filter_key[category]['store-mode'] == 'v':
                                    
                        # is the filter a DICT
                        if filter_key[category]['value-dict']:

                            # then correct match needed
                            if filter.lower() == card[filter_key[category]['section']][category].lower():
                                fits = True
                                break

                        # NOT dict
                        else:
                                                
                            if operation == 'gte':
 
                                # then >= needed
                                if self.to_integer(filter) <= self.to_integer(card[filter_key[category]['section']][category]):
                                    fits = True
                                    break                                                

                            elif operation == 'lte': 
                                                
                                # then <= needed
                                if self.to_integer(filter) >= self.to_integer(card[filter_key[category]['section']][category]):
                                    fits = True
                                    break                                                     
                                                
                            #NOT correct mach needed
                            elif filter.lower() in card[filter_key[category]['section']][category].lower():
                                fits = True
                                break
                                    
                    elif filter_key[category]['store-mode'] == 't':
                                        
                        # is the filter a DICT
                        if filter_key[category]['value-dict']:

                            # then correct match needed
                            if filter.lower() == card[filter_key[category]['section']][config_ini['language']].lower():
                                fits = True
                                break
                                        
                        # NOT dict
                        else:
                                            
                            if filter.lower() in card[filter_key[category]['section']][config_ini['language']].lower():
                                fits = True
                                break                                
                                
                    # if at least one filter matches to a category values in the card
                    if fits:
                        break

            if not fits:
                break                
        
        return fits
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
    # ----------------------------
    #
    # Key Press Event: Enter/Space
    #
    # ----------------------------
    def keyPressEvent(self, event):

        #
        # Enter / Return / Space
        #
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter or event.key() == Qt.Key_Space:

            # Simulate a Mouse Press / Release Event on the Image
            if self.actual_card_holder.shown_card_list and len(self.actual_card_holder.shown_card_list) > 0:
                card=self.actual_card_holder.shown_card_list[0]
                if card.is_selected():
                    
                    event_press = QMouseEvent(QEvent.MouseButtonPress, QPoint(10,10), Qt.LeftButton, Qt.LeftButton, Qt.NoModifier)
                    event_release = QMouseEvent(QEvent.MouseButtonRelease, QPoint(10,10), Qt.LeftButton, Qt.LeftButton, Qt.NoModifier)

                    layout=card.get_panel().get_layout()                    
                    card_panel = layout.itemAt(0).widget()
                    card_panel.card_image.mousePressEvent(event_press)
                    card_panel.card_image.mouseReleaseEvent(event_release)            

        #
        # Escape
        #
        if event.key() == Qt.Key_Escape:
            self.restore_previous_holder()

        event.ignore()
  

class LinkLabel(QLabel):
    def __init__(self, text, parent, index):
        QLabel.__init__(self, text, parent)
        self.parent = parent
        self.index = index        
        self.setFont(QFont( FONT_TYPE, HIERARCHY_TITLE_FONT_SIZE, weight=QFont.Bold if index else QFont.Normal))

    # Mouse Hover in
    def enterEvent(self, event):
        if self.index:
            QApplication.setOverrideCursor(Qt.PointingHandCursor)
        event.ignore()

    # Mouse Hover out
    def leaveEvent(self, event):
        if self.index:
            QApplication.restoreOverrideCursor()
        event.ignore()

    # Mouse Press
    def mousePressEvent(self, event):
        if event.buttons() == Qt.LeftButton and self.index:
            self.parent.panel.restore_previous_holder(self.index)
        event.ignore()
    

# =========================================
# 
# This Class represents the title
#
# =========================================
#
class HierarchyTitle(QWidget):
    DEFAULT_BACKGROUND_COLOR = Qt.lightGray
    DEFAULT_BORDER_RADIUS = 10
    
    def __init__(self, parent, panel):
        QWidget.__init__(self, parent)

        self.panel = panel
        
        self_layout = QHBoxLayout(self)
        self_layout.setContentsMargins(5, 5, 5, 5)
        self_layout.setSpacing(0)
        #self_layout.setAlignment(Qt.AlignHCenter)
        self.setLayout(self_layout)

        self.text = QWidget(self)
        #self.text.setWordWrap(True)
        #self.text.setAlignment(Qt.AlignHCenter)
        #self.text.setFont(QFont( FONT_TYPE, HIERARCHY_TITLE_FONT_SIZE, weight=QFont.Bold))
        self_layout.addWidget(self.text)
        
        #self.text_layout = FlowLayout(self.text)
        self.text_layout = QVBoxLayout(self.text)
        self.text_layout.setContentsMargins(0, 0, 0, 0)
        self.text_layout.setSpacing(0)
        self.text.setLayout(self.text_layout)

        self.set_background_color(QColor(HierarchyTitle.DEFAULT_BACKGROUND_COLOR), False)
        self.set_border_radius(HierarchyTitle.DEFAULT_BORDER_RADIUS, False)
        
        self.card_holder_history = None
        self.actual_card_holder = None
        self.no_refresh = False
        
#    def resizeEvent(self, event):
#        if not self.no_refresh:
#            self.refresh_title()
#        event.accept()
    
    def refresh_title(self):
        if self.card_holder_history and self.actual_card_holder:
            self.set_title(self.card_holder_history, self.actual_card_holder)
    
    def set_title(self, card_holder_history, actual_card_holder):
        #self.blockSignals(True)
        #self.no_refresh = True

        clearLayout(self.text_layout)
 
        self.card_holder_history = card_holder_history
        self.actual_card_holder = actual_card_holder
 
        history = []
        for index, card in enumerate(card_holder_history):
            if card.title:
                label = LinkLabel(card.title, self, len(card_holder_history)-index)
                history.append(label)
       
        minimumWidth = 0
        max_width = self.size().width() - 2 * 5        
        self.create_one_line_container()
        
        for cw in history:
            
            minimumWidth += self.get_width_in_pixels(cw)
            if minimumWidth <= max_width:            
                self.add_to_one_line_container(cw)
            else:
                self.push_new_line_container()
                self.create_one_line_container()
                minimumWidth = 0
                self.add_to_one_line_container(cw)
            
            separator = QLabel(' > ')
            separator.setFont(QFont( FONT_TYPE, HIERARCHY_TITLE_FONT_SIZE, weight=QFont.Normal ))
            minimumWidth += self.get_width_in_pixels(separator)
            if minimumWidth <= max_width:            
                self.add_to_one_line_container(separator)
            else:
                self.push_new_line_container()
                self.create_one_line_container()
                minimumWidth = 0
                self.add_to_one_line_container(separator)
                                       
        notLinkTitle = LinkLabel(actual_card_holder.title, self, 0)
        minimumWidth += self.get_width_in_pixels(notLinkTitle)
        if minimumWidth <= max_width:            
            self.add_to_one_line_container(notLinkTitle)
        else:
            self.push_new_line_container()
            self.create_one_line_container()
            minimumWidth = 0
            self.add_to_one_line_container(notLinkTitle)
        self.push_new_line_container()
       



    def create_one_line_container(self):
        self.one_line_container = QWidget(self)
        self.one_line_container_layout = QHBoxLayout(self.one_line_container)
        self.one_line_container_layout.setContentsMargins(0, 0, 0, 0)
        self.one_line_container_layout.setSpacing(0)
        self.one_line_container.setLayout(self.one_line_container_layout)
        self.one_line_container_layout.setAlignment(Qt.AlignHCenter)

    def add_to_one_line_container(self, cw):
        self.one_line_container_layout.addWidget(cw)
        
    def push_new_line_container(self):
        self.text_layout.addWidget(self.one_line_container)
        
    def get_width_in_pixels(self, cw):
        initialRect = cw.fontMetrics().boundingRect(cw.text());
        improvedRect = cw.fontMetrics().boundingRect(initialRect, 0, cw.text());   
        return improvedRect.width()
        
        
    def set_background_color(self, color, update=False):
        self.background_color = color
        self.text.setStyleSheet('background: ' + color.name()) 
        if update:
            self.update()
            
    def set_border_radius(self, radius, update=True):
        self.border_radius = radius
        if update:
            self.update()            
        
    def paintEvent(self, event):
        s = self.size()
        qp = QPainter()
        qp.begin(self)
        qp.setRenderHint(QPainter.Antialiasing, True)
        qp.setBrush( self.background_color )

        qp.drawRoundedRect(0, 0, s.width(), s.height(), self.border_radius, self.border_radius)
        qp.end()
        





        

# =========================================
#
#          Control Panel 
#
# on the TOP of the Window
#
# Contains:
#           Button Control
#           Fast Search Control
#           Advanced Search Control
#
# =========================================
class ControlPanel(QWidget):
    def __init__(self, gui):
        super().__init__(gui)
       
        self.gui = gui
        
        self_layout = QVBoxLayout(self)
        self.setLayout(self_layout)
        
        # controls the distance between the MainWindow and the added container: scrollContent
        self_layout.setContentsMargins(3, 3, 3, 3)
        self_layout.setSpacing(5)

        # ----------------------
        #
        # Control Buttons Holder
        #
        # ----------------------
        self.control_buttons_holder = ControlButtonsHolder(self)
        self_layout.addWidget(self.control_buttons_holder)
        
        # ------------------
        #
        # Fast Filter Holder
        #
        # ------------------
        self.fast_filter_holder = FastFilterHolder()
        self.fast_filter_holder.changed.connect(self.fast_filter_on_change)
        self.fast_filter_holder.clear_clicked.connect(self.fast_filter_clear_on_change)
        self.fast_filter_holder.setHidden(True)        
        self_layout.addWidget(self.fast_filter_holder)

        # ----------------------
        #
        # Advanced Filter Holder
        #
        # ----------------------
        self.advanced_filter_holder = AdvancedFilterHolder(self)
        self.advanced_filter_holder.filter_clicked.connect(self.advanced_filter_filter_on_click)
        self.advanced_filter_holder.clear_clicked.connect(self.advanced_filter_clear_on_click)
        self.advanced_filter_holder.setHidden(True)
        self_layout.addWidget(self.advanced_filter_holder)

        # Listeners
        self.back_button_listener = None
        self.fast_filter_listener = None
        self.advanced_filter_listener = None

    def refresh_label(self):
        self.fast_filter_holder.refresh_label()
        self.advanced_filter_holder.refresh_label()

    def set_back_button_method(self, method):
        self.control_buttons_holder.back_button_method = method
        
    def set_fast_filter_listener(self, listener):
        self.fast_filter_listener = listener
 
    # ----------------------------------
    #
    # a value changed in the fast filter
    # 
    # ----------------------------------
    def fast_filter_on_change(self):
        """
        --- Fast Filter on Change ---
        
        If there is no listener then Filters the Cards
        """
        if self.fast_filter_listener:
            self.gui.filter_the_cards()
            
    def fast_filter_clear_on_change(self):
        """
        --- Fast Filter Clear on Change ---
        
        -Clear all the fields
        -Filter the cards
        """
        self.fast_filter_listener = None
        self.fast_filter_holder.clear_fields()
        self.fast_filter_listener = self.gui
        self.fast_filter_on_change()
    
    def get_fast_filter_holder(self):
        return self.fast_filter_holder

    # ---------------------------------
    #
    # advanced filter clicked
    #
    # ---------------------------------
    def advanced_filter_filter_on_click(self):
        """
        --- Advanced Filter FILTER on Click
        
        The Filter button was clicked on the Advanced Filter panel
        -filter the cards
        """
        self.gui.filter_the_cards()
        
    def advanced_filter_clear_on_click(self):
        """
        --- Advanced Filter CLEAR on Click ---
        
        The CLEAR button was clicked on the Advanced Filter panel
        -clear the fields
        -filter the cards by cleared filter
        """
        self.advanced_filter_holder.clear_fields()
        self.gui.filter_the_cards()

    def get_advanced_filter_holder(self):
        return self.advanced_filter_holder

        
def main():    
    app = QApplication(sys.argv)
    
    ex = GuiAkoTeka()
    ex.startCardHolder()
    sys.exit(app.exec_())
    
    



import os
import re
import configparser

from akoteka.handle_property import get_config_ini
from akoteka.handle_property import config_ini
from akoteka.handle_property import _

from akoteka.accessories import collect_cards
from akoteka.accessories import get_pattern_video
from akoteka.accessories import get_pattern_audio
from akoteka.accessories import get_pattern_image
from akoteka.accessories import get_pattern_card
from akoteka.accessories import get_pattern_length
from akoteka.accessories import get_pattern_year

HIGHLIGHT = '\033[31m'
COLORBACK = '\033[0;0m'

#def folder_investigation( actual_dir, json_list):
def folder_investigation( actual_dir):
    
    # Collect files and and dirs in the current directory
    file_list = [f for f in os.listdir(actual_dir) if os.path.isfile(os.path.join(actual_dir, f))] if os.path.exists(actual_dir) else []
    dir_list = [d for d in os.listdir(actual_dir) if os.path.isdir(os.path.join(actual_dir, d))] if os.path.exists(actual_dir) else []

    # now I got to a certain level of directory structure
    card_path_os = None
    media_path_os = None
    image_path_os = None
    media_name = None
    
    is_card_dir = True

    # Go through all files in the folder
    for file_name in file_list:

        # find the Card
        if get_pattern_card().match( file_name ):
            card_path_os = os.path.join(actual_dir, file_name)
            
        # find the Media (video or audio)
        if get_pattern_audio().match(file_name) or get_pattern_video().match(file_name):
            media_path_os = os.path.join(actual_dir, file_name)
            media_name = file_name
            
        # find the Image
        if get_pattern_image().match( file_name ):
           image_path_os = os.path.join(actual_dir, file_name)

    card = {}
    
    card['title'] = {}
    card['storyline'] = {}
                
    general_json_list = {}
    general_json_list['director'] = []
    general_json_list['sound'] = []
    general_json_list['sub'] = [] 
    general_json_list['genre'] = []
    general_json_list['theme'] = []
    general_json_list['actor'] = []
    general_json_list['country'] = []
    card['general'] = general_json_list
                
    card['rating'] = {}                                        
    card['links'] = {}

    extra_json_list = {}    
    extra_json_list['sub-cards'] = []
    extra_json_list['recent-folder'] = actual_dir
    card['extra'] = extra_json_list

    # ----------------------------------
    #
    # it is a COLLECTOR CARD dir
    #
    # there is:     -Card 
    #               -at least one Dir
    # ther is NO:   -Media
    #  
    # ----------------------------------
    if card_path_os and not media_path_os and dir_list:
                
        parser = configparser.RawConfigParser()
        parser.read(card_path_os, encoding='utf-8')
        
        try:            
            # save the http path of the image
            card['extra']['image-path'] = image_path_os

            # saves the os path of the media - There is no
            card['extra']['media-path'] = None
            
            card['title']['hu'] = parser.get("titles", "title_hu")
            card['title']['en'] = parser.get("titles", "title_en")
                
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            print(nop_err, "in ", card_path_os)
        
    # --------------------------------
    #
    # it is a MEDIA CARD dir
    #
    # there is:     -Card 
    #               -Media
    # 
    # --------------------------------
    elif card_path_os and media_path_os:

        err = []

        # first collect every data from the card
        parser = configparser.RawConfigParser()
        parser.read(card_path_os, encoding='utf-8')

        # save the os path of the image
        card['extra']['image-path'] = image_path_os            

        # saves the os path of the media
        card['extra']['media-path'] = media_path_os

        try:
            card['title']['hu'] = parser.get("titles", "title_hu")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['title']['hu']:
                err.append("Empty attribute: 'title_hu' in section: 'titles'")
                
        try:            
            card['title']['en'] = parser.get("titles", "title_en")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['title']['en']:
                err.append("Empty attribute: 'title_en' in section: 'titles'")

        try:
            card['storyline']['hu'] = parser.get("storyline", "storyline_hu")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))

        try:
            card['storyline']['en'] = parser.get("storyline", "storyline_en")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))

        try:            
            card['general']['year'] = parser.get("general", "year")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['general']['year'] :
                err.append("Empty attribute: 'year' in section: 'general section'")
            elif not get_pattern_year().match( card['general']['year'] ):
                err.append("Wrong attribute: 'year' in section: 'general' value: " + HIGHLIGHT + card['general']['year'] + COLORBACK)
            
        try:
            directors = parser.get("general", "director").split(",")            
            for director in directors:
                card['general']['director'].append(director.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['general']['director']:
                err.append("Empty attribute: 'director' in section: 'general'")
            
        try:
            card['general']['length'] = parser.get("general", "length")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['general']['length']:
                err.append("Empty attribute: 'length' in section: 'general'")
            elif not get_pattern_length().match( card['general']['length'] ):
                err.append("Wrong attribute: 'length' in section: 'general' value: " + HIGHLIGHT + card['general']['length'] + COLORBACK)
                
        try:
            sounds = parser.get("general", "sound").split(",")
            for sound in sounds:
                if sound:
                    if "lang_" in _("lang_" + sound.strip()):
                        err.append("Missing translation for 'sound': " + HIGHLIGHT + "lang_" + sound.strip() + COLORBACK + " in property files.")
                    if "lang_long_" in _("lang_long_" + sound.strip()):
                        err.append("Missing translation for 'sound': " + HIGHLIGHT + "lang_long_" + sound.strip() + COLORBACK + " in property files.")                    
                    card['general']['sound'].append(sound.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['general']['sound']:
                err.append("Empty attribute: 'sound' in section: 'general'")

        try:
            subs = parser.get("general", "sub").split(",")
            for sub in subs:
                if sub:
                    if "lang_" in _("lang_" + sub.strip()):
                        err.append("Missing translation for 'sub': " + HIGHLIGHT + "lang_" + sub.strip() + COLORBACK + " in property files.")
                    if "lang_long_" in _("lang_long_" + sub.strip()):
                        err.append("Missing translation for 'sub': " + HIGHLIGHT + "lang_long_" + sub.strip() + COLORBACK + " in property files.")                    
                    card['general']['sub'].append(sub.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        
        try:
            genres = parser.get("general", "genre").split(",")
            for genre in genres:
                if "genre_" in _("genre_" + genre.strip()):
                    err.append("Missing translation for 'genre': " + HIGHLIGHT + "genre_" + genre.strip() + COLORBACK + " in property files.")
                card['general']['genre'].append(genre.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['general']['genre']:
                err.append("Empty attribute: 'genre' in section: 'general'")

        try:
            themes = parser.get("general", "theme").split(",")
            for theme in themes:
                if theme:
                    if "theme_" in _("theme_" + theme.strip()):
                        err.append("Missing translation for 'theme': " + HIGHLIGHT + "theme_" + theme.strip() + COLORBACK + " in property files.")                
                    card['general']['theme'].append(theme.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))

        try:
            actors = parser.get("general", "actor").split(",")
            for actor in actors:
                card['general']['actor'].append(actor.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))

        try:
            countries = parser.get("general", "country").split(",")
            for country in countries:
                if "country_" in _("country_" + country.strip()):
                    err.append("Missing translation: " + HIGHLIGHT + "country_" + country.strip() + COLORBACK + " in property files.")
                if "country_long_" in _("country_long_" + country.strip()):
                    err.append("Missing translation: " + HIGHLIGHT + "country_long_" + country.strip() + COLORBACK + " in property files.")                   
                card['general']['country'].append(country.strip())
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['general']['country']:
                err.append("Empty attribute: 'country' in section: 'general'")

        try:
            card['rating']['best'] = parser.get("rating", "best")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['rating']['best']:
                err.append("Empty attribute: 'best' in section: 'rating'")
            
        try:
            card['rating']['new'] = parser.get("rating", "new")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['rating']['new']:
                err.append("Empty attribute: 'new' in section: 'rating'")

        try:
            card['rating']['favorite'] = parser.get("rating", "favorite")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))
        else:
            if not card['rating']['favorite']:
                err.append("Empty attribute: 'favorite' in section: 'rating'")

        try:                                                
            card['links']['imdb'] = parser.get("links", "imdb")
        except (configparser.NoSectionError, configparser.NoOptionError) as nop_err:
            err.append(str(nop_err))        

        if err:
            print(card_path_os)
            for msg in err:
                print("       " + msg)

    # ----------------------------------
    #
    # it is NO CARD dir
    #
    # ----------------------------------
    else:
        
        is_card_dir = False
        
    # ----------------------------
    #
    # Through the Sub-directories
    #
    # ----------------------------    
    for name in dir_list:
        subfolder_path_os = os.path.join(actual_dir, name)
        folder_investigation( subfolder_path_os )

    # and finaly returns
    return

def main():    
    config_ini_function = get_config_ini()
    paths = config_ini_function.get_media_path()
    folder_investigation(paths)



import sys
import os
import importlib
from threading import Thread
from pkg_resources import resource_string, resource_filename
from functools import cmp_to_key
import locale

from cardholder.cardholder import CardHolder
from cardholder.cardholder import Card
from cardholder.cardholder import CollectCardsThread

from aqfilter.aqfilter import AQFilter

from akoteka.gui.pyqt_import import *
from akoteka.gui.card_panel import CardPanel
from akoteka.gui.configuration_dialog import ConfigurationDialog
from akoteka.gui.control_buttons_holder import ControlButtonsHolder

from akoteka.accessories import collect_cards
from akoteka.accessories import filter_key
from akoteka.accessories import clearLayout
from akoteka.accessories import FlowLayout

from akoteka.constants import *
from akoteka.setup.setup import getSetupIni

from akoteka.handle_property import _
from akoteka.handle_property import re_read_config_ini
from akoteka.handle_property import config_ini
from akoteka.handle_property import get_config_ini

# =======================
#
# Advanced Filter HOLDER
#
# =======================
class AdvancedFilterHolder(QWidget):
    
    filter_clicked = pyqtSignal()
    clear_clicked = pyqtSignal()
    
    def __init__(self, parent):
        super().__init__(parent)

        self_layout = QVBoxLayout(self)
        self.setLayout( self_layout )
        self_layout.setContentsMargins(0, 0, 0, 0)
        self_layout.setSpacing(0)    

        holder = QWidget(self)
        holder_layout = QGridLayout(holder)
        holder.setLayout( holder_layout )
        holder_layout.setContentsMargins(0, 0, 0, 0)
        holder_layout.setSpacing(1)

        self_layout.addWidget(QHLine())
        self_layout.addWidget(holder)
        
        filter_style_box = '''
            AQFilter { 
                max-width: 200px; min-width: 200px; border: 1px solid gray; border-radius: 5px;
            }
        '''

        combobox_short_style_box = '''
            QComboBox { 
                min-width: 40px; border: 1px solid gray; border-radius: 5px;
            }
        '''
        combobox_long_style_box = '''
            QComboBox { 
                min-width: 60px; border: 1px solid gray; border-radius: 5px;
            }
        '''
        
        dropdown_style_box ='''
            QComboBox QAbstractItemView::item { 
                color: red;
                max-height: 15px;
            }
        '''            

        # ----------
        #
        # Title
        #
        # ----------
        self.title_label = QLabel(_('title_title') + ": ")
        self.title_filter = AQFilter(holder)
        self.title_filter.setStyleSheet(filter_style_box)
        holder_layout.addWidget(self.title_label, 0, 0, 1, 1)
        holder_layout.addWidget(self.title_filter, 0, 1, 1, 1)

        # ----------
        #
        # Director
        #
        # ----------
        self.director_label = QLabel(_('title_director') + ": ")
        self.director_filter = AQFilter(holder)
        self.director_filter.setStyleSheet(filter_style_box)
        #self.director_filter.setMinCharsToShowList(0)
        holder_layout.addWidget(self.director_label, 1, 0, 1, 1)
        holder_layout.addWidget(self.director_filter, 1, 1, 1, 1)
        
        # ----------
        #
        # Actors
        #
        # ----------
        self.actor_label = QLabel(_('title_actor') + ": ")
        self.actor_filter = AQFilter(holder)
        self.actor_filter.setStyleSheet(filter_style_box)
        holder_layout.addWidget(self.actor_label, 2, 0, 1, 1)
        holder_layout.addWidget(self.actor_filter, 2, 1, 1, 1)

        # ---
     
        # ----------
        #
        # Genre
        #
        # ----------
        self.genre_label = QLabel(_('title_genre') + ": ")
        self.genre_filter = AQFilter(holder)
        self.genre_filter.setStyleSheet(filter_style_box)
        holder_layout.addWidget(self.genre_label, 0, 2, 1, 1)
        holder_layout.addWidget(self.genre_filter, 0, 3, 1, 1)
        
        # ----------
        #
        # Theme
        #
        # ----------
        self.theme_label = QLabel(_('title_theme') + ": ")
        self.theme_filter = AQFilter(holder)
        self.theme_filter.setStyleSheet(filter_style_box)
        holder_layout.addWidget(self.theme_label, 1, 2, 1, 1)
        holder_layout.addWidget(self.theme_filter, 1, 3, 1, 1)

        # ---
        
        # ----------
        #
        # Sound
        #
        # ----------
        self.sound_label = QLabel(_('title_sound') + ": ")
        self.sound_combobox = QComboBox(self)
        self.sound_combobox.setFocusPolicy(Qt.NoFocus)
        self.sound_combobox.setStyleSheet(combobox_long_style_box + dropdown_style_box)
        self.sound_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
  
        holder_layout.addWidget(self.sound_label, 0, 4, 1, 1)
        holder_layout.addWidget(self.sound_combobox, 0, 5, 1, 1)

        # ----------
        #
        # Subtitle
        #
        # ----------
        self.sub_label = QLabel(_('title_sub') + ": ")
        self.sub_combobox = QComboBox(self)
        self.sub_combobox.setFocusPolicy(Qt.NoFocus)
        self.sub_combobox.setStyleSheet(combobox_long_style_box + dropdown_style_box)
        self.sub_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
  
        holder_layout.addWidget(self.sub_label, 1, 4, 1, 1)
        holder_layout.addWidget(self.sub_combobox, 1, 5, 1, 1)
  
        # ----------
        #
        # Country
        #
        # ----------
        self.country_label = QLabel(_('title_country') + ": ")
        self.country_combobox = QComboBox(self)
        self.country_combobox.setFocusPolicy(Qt.NoFocus)
        self.country_combobox.setStyleSheet(combobox_long_style_box + dropdown_style_box)        
        self.country_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
  
        holder_layout.addWidget(self.country_label, 2, 4, 1, 1)
        holder_layout.addWidget(self.country_combobox, 2, 5, 1, 1)

        # ---
        
        # ----------
        #
        # Year
        #
        # ----------
        self.year_from_label = QLabel(_('title_year') + ": ")
        year_to_label = QLabel('-')
        self.year_from_combobox = QComboBox(self)
        self.year_from_combobox.setFocusPolicy(Qt.NoFocus)
        self.year_from_combobox.setStyleSheet(combobox_short_style_box + dropdown_style_box)
        self.year_from_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        #self.year_from_combobox.addItem("        ")

        self.year_to_combobox = QComboBox()
        self.year_to_combobox.setFocusPolicy(Qt.NoFocus)
        self.year_to_combobox.setStyleSheet(combobox_short_style_box + dropdown_style_box)
        self.year_to_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        #self.year_to_combobox.addItem("        ")
  
        holder_layout.addWidget(self.year_from_label, 0, 6, 1, 1)
        holder_layout.addWidget(self.year_from_combobox, 0, 7, 1, 1)
        holder_layout.addWidget(year_to_label, 0, 8, 1, 1)
        holder_layout.addWidget(self.year_to_combobox, 0, 9, 1, 1)

        # ----------
        #
        # Length
        #
        # ----------
        self.length_from_label = QLabel(_('title_length') + ": ")
        length_to_label = QLabel('-')
        self.length_from_combobox = QComboBox()
        self.length_from_combobox.setFocusPolicy(Qt.NoFocus)
        self.length_from_combobox.setStyleSheet(combobox_short_style_box + dropdown_style_box)
        self.length_from_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        #self.length_from_combobox.addItem("        ")

        self.length_to_combobox = QComboBox()
        self.length_to_combobox.setFocusPolicy(Qt.NoFocus)
        self.length_to_combobox.setStyleSheet(combobox_short_style_box + dropdown_style_box)        
        self.length_to_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        #self.length_to_combobox.addItem("        ")
  
        holder_layout.addWidget(self.length_from_label, 1, 6, 1, 1)
        holder_layout.addWidget(self.length_from_combobox, 1, 7, 1, 1)
        holder_layout.addWidget(length_to_label, 1, 8, 1, 1)
        holder_layout.addWidget(self.length_to_combobox, 1, 9, 1, 1)

        # ------------------
        #
        # Vertical Separator
        #
        # ------------------
        holder_layout.setColumnStretch(10, 1)
        holder_layout.addWidget(QVLine(), 0, 11, 3, 2)
        holder_layout.setColumnStretch(12, 1)

        # -------------
        #
        # Clear button
        #
        # -------------
        self.clear_button = QPushButton(_('button_clear'))
        holder_layout.addWidget(self.clear_button, 0, 13, 1, 1)
        self.clear_button.clicked.connect(self.clear_button_clicked)
        
        # -------------
        #
        # Search button
        #
        # -------------
        self.filter_button = QPushButton(_('button_filter'))
        holder_layout.addWidget(self.filter_button, 2, 13, 1, 1)
        self.filter_button.clicked.connect(self.filter_button_clicked)

        # ----------
        #
        # Stretch
        #
        # ----------
        holder_layout.setColumnStretch(11, 1)

    def refresh_label(self):
       self.title_label.setText(_('title_title') + ": ")
       self.director_label.setText(_('title_director') + ": ")
       self.actor_label.setText(_('title_actor') + ": ")
       self.genre_label.setText(_('title_genre') + ": ")
       self.theme_label.setText(_('title_theme') + ": ")
       self.sound_label.setText(_('title_sound') + ": ")
       self.sub_label.setText(_('title_sub') + ": ")
       self.country_label.setText(_('title_country') + ": ")
       self.year_from_label.setText(_('title_year') + ": ")
       self.length_from_label.setText(_('title_length') + ": ")
       
    def clear_title(self):
        self.title_filter.clear()

    def clear_director(self):
        self.director_filter.clear()

    def clear_actor(self):
        self.actor_filter.clear()

    def clear_genre(self):
        self.genre_filter.clear()

    def clear_theme(self):
        self.theme_filter.clear()

    def clear_sound(self):
        self.sound_combobox.clear()
        self.sound_combobox.addItem("")
        
    def clear_sub(self):
        self.sub_combobox.clear()
        self.sub_combobox.addItem("")
        
    def clear_country(self):
        self.country_combobox.clear()
        self.country_combobox.addItem("")

    def clear_year(self):
        self.year_from_combobox.clear()
        self.year_to_combobox.clear()
        self.year_from_combobox.addItem("")
        self.year_to_combobox.addItem("")

    def clear_length(self):
        self.length_from_combobox.clear()
        self.length_to_combobox.clear()
        self.length_from_combobox.addItem("")
        self.length_to_combobox.addItem("")

    def clear_fields(self):
        self.clear_title()
        self.clear_director()
        self.clear_actor()
        self.clear_genre()
        self.clear_theme()
        self.clear_sound()
        self.clear_sub()
        self.clear_country()
        self.clear_year()
        self.clear_length()

    # ---
    
    def set_title(self, value):
        self.title_filter.setValue(value)

    def set_genre(self, value):
        self.genre_filter.setValue(value)
    
    def set_theme(self, value):
        self.theme_filter.setValue(value)
        
    def set_director(self, value):
        self.director_filter.setValue(value)
        
    def set_actor(self, value):
        self.actor_filter.setValue(value)

    def select_sound(self, id):
        self.sound_combobox.setCurrentIndex( self.sound_combobox.findText(id) )

    def select_sub(self, id):
        self.sub_combobox.setCurrentIndex( self.sub_combobox.findText(id) )

    def select_country(self, id):
        self.country_combobox.setCurrentIndex( self.country_combobox.findText(id) )

    def select_year_from(self, value):
        self.year_from_combobox.setCurrentIndex( self.year_from_combobox.findText(value) )

    def select_year_to(self, value):
        self.year_to_combobox.setCurrentIndex( self.year_to_combobox.findText(value) )

    def select_length_from(self, value):
        self.length_from_combobox.setCurrentIndex( self.length_from_combobox.findText(value) )

    def select_length_to(self, value):
        self.length_to_combobox.setCurrentIndex( self.length_to_combobox.findText(value) )
    
    # ---
    
    def add_title(self, value):
        self.title_filter.addItemToList(value, value)
    
    def add_director(self, director):
        self.director_filter.addItemToList(director, director)

    def add_actor(self, actor):
        self.actor_filter.addItemToList(actor, actor)

    def add_genre(self, genre, index):
        self.genre_filter.addItemToList(genre, index)
    
    def add_theme(self, theme, index):
        self.theme_filter.addItemToList(theme, index)
    
    def add_sound(self, value, id):
        self.sound_combobox.addItem(value, id)
    
    def add_sub(self, value, id):
        self.sub_combobox.addItem(value, id)
 
    def add_country(self, value, id):
        self.country_combobox.addItem(value, id)
        
    def add_length(self, value):
        self.length_from_combobox.addItem(value, value)
        self.length_to_combobox.addItem(value, value)

    def add_year(self, value):
        self.year_from_combobox.addItem(value, value)
        self.year_to_combobox.addItem(value, value)
        
    # ---
    
    def get_sound_selected_index(self):
        return self.sound_combobox.itemData( self.sound_combobox.currentIndex() )

    def get_sound_selected_value(self):
        return self.sound_combobox.itemText( self.sound_combobox.currentIndex() )

    def get_sub_selected_index(self):
        return self.sub_combobox.itemData( self.sub_combobox.currentIndex() )

    def get_sub_selected_value(self):
        return self.sub_combobox.itemText( self.sub_combobox.currentIndex() )

    def get_country_selected_index(self):
        return self.country_combobox.itemData( self.country_combobox.currentIndex() )

    def get_country_selected_value(self):
        return self.country_combobox.itemText( self.country_combobox.currentIndex() )

    def get_length_from_selected_value(self):
        return self.length_from_combobox.currentText()

    def get_length_to_selected_value(self):
        return self.length_to_combobox.currentText()

    def get_year_from_selected_value(self):
        return self.year_from_combobox.currentText()

    def get_year_to_selected_value(self):
        return self.year_to_combobox.currentText()

    # ---
    
    def filter_button_clicked(self):
        self.filter_clicked.emit()        

    def clear_button_clicked(self):
        self.clear_clicked.emit()        

    def get_filter_selection(self):
        
        # -----------------------------------------------------------------
        #
        # 1st value:    the typed value
        # 2nd value:    
        #               if not from DICT:   None
        #               if from DICT:
        #                               if found: [found, dict, elements]
        #                               if not found: []
        #
        # -----------------------------------------------------------------
        filter_selection = {
            "title":    ["title", self.title_filter.getValue(), None, "a"],
            "genre":    ["genre", self.genre_filter.getValue(), self.genre_filter.getIndexes(), "a"],
            "theme":    ["theme", self.theme_filter.getValue(), self.theme_filter.getIndexes(), "a"],
            "director": ["director", self.director_filter.getValue(), None, "a"],
            "actor":    ["actor", self.actor_filter.getValue(), None, "a"],
            "sound":    ["sound", self.get_sound_selected_value(), [self.get_sound_selected_index()], "a"],
            "sub":      ["sub", self.get_sub_selected_value(), [self.get_sub_selected_index()], "a"],
            "country":  ["country", self.get_country_selected_value(), [self.get_country_selected_index()], "a"],
            "length-from":  ["length", self.get_length_from_selected_value(), None, "gte"],
            "length-to":    ["length", self.get_length_to_selected_value(), None, "lte"],
            "year-from":    ["year", self.get_year_from_selected_value(), None, "gte"],
            "year-to":  ["year", self.get_year_to_selected_value(), None, "lte"],
        }
        return filter_selection

    
    

        




